/*++

	Copyright (c) 2008-2009 SMSC. All rights reserved.

	The information contained in this file is confidential and proprietary to
	Standard Microsystems Corporation.	No part of this file may be reproduced
	or distributed, in any form or by any means for any purpose, without the
	express written permission of Standard Microsystems Corporation.

	
Module Name:

	interrupt.c
		
Abstract:

	This module contains miniport functions for handling interrupt
	transactions and other helper routines called by these miniport 
	functions.
	
Revision History:
	12/11/2008	Initial version.

Notes:

--*/


#include "smsc9500.h"

#ifdef USE_INT_PIPE

/*++

Routine Description:

   Workitem to post all the free int requests to the target
   driver. This workitem is scheduled from the MiniportInitialize
   worker routine during start and from the NICFreeICB routine
   whenever the outstanding int requests to the target driver goes 
   below the watermark.
	  
Arguments:

	WorkItem - Pointer to workitem
	
	Context - Pointer to our Adapter

Return Value:

	None.

--*/
VOID
NICPostIntsWorkItemCallBack(
	PNDIS_WORK_ITEM WorkItem, 
	PVOID Context
	)

{
	PMP_ADAPTER 	Adapter = (PMP_ADAPTER)Context;
	NTSTATUS		ntStatus;
	PICB			pICB = NULL;

	SMSC_TRACE0(DBG_INT,"--> NICPostIntsWorkItemCallBack\r\n"); 
	
	//
	// Clear the flag to let the WorkItem structure be reused for
	// scheduling another one. Clear it before processing to avoid 
	// race condition which can leave the device in a state where
	// no further int transactions are submitted
	//
	InterlockedExchange(&Adapter->IsPostIntsWorkItemQueued, FALSE);

	// Now lets actually do the interrupt submissions
	NdisAcquireSpinLock(&Adapter->IntLock);

	while(MP_IS_READY(Adapter) && !IsListEmpty(&Adapter->IntFreeList) && MP_TEST_FLAG(Adapter, fMP_POST_INTS))
	{
		pICB = (PICB) RemoveHeadList(&Adapter->IntFreeList);

		ASSERT(pICB);// cannot be NULL
		
		//
		// Insert the ICB in the Int busy queue
		//	   
		InsertTailList(&Adapter->IntBusyList, &pICB->List);
		NdisInterlockedIncrement(&Adapter->nBusyInt);		   
		ASSERT(Adapter->nBusyInt <= Adapter->MaxICBs);
		
		NdisReleaseSpinLock(&Adapter->IntLock);

		Adapter->nIntsPosted++;
		ntStatus = NICPostIntRequest(Adapter, pICB);
		
		if (!NT_SUCCESS ( ntStatus ) ) 
		{
			SMSC_TRACE1(DBG_ERROR, "NICPostIntRequest failed %x\r\n", ntStatus);

			// NICFreeICB expects it in the Processed queue, so put it there	
			// before making the call.
			NdisAcquireSpinLock(&Adapter->IntLock);

			// Remove it from the busy list 
			RemoveEntryList(&pICB->List);	
			NdisInterlockedDecrement(&Adapter->nBusyInt);
			ASSERT(Adapter->nBusyInt >= 0);

			// Put it in the Processed list.
			InsertTailList(&Adapter->IntProcessedList, &pICB->List);
			NdisInterlockedIncrement(&Adapter->nProcessedInt);
			ASSERT(Adapter->nProcessedInt <= Adapter->MaxICBs);

			SMSC_TRACE1(DBG_INT,"ICB %p removed from the busy and inserted into IntProcessedList\r\n", pICB);

			NdisReleaseSpinLock(&Adapter->IntLock); 

			if (NdisInterlockedDecrement(&pICB->Ref) == 0) 
			{
				NICFreeICB(pICB);
			}
			else
			{
				// This should never happen
				SMSC_TRACE0(DBG_ERROR, "NICPostIntsWorkItemCallBack finds ICB with invalid Ref count!\r\n");			  
			}

			NdisAcquireSpinLock(&Adapter->IntLock);
			break;			  
		}
		
		NdisAcquireSpinLock(&Adapter->IntLock);
		
	}

	NdisReleaseSpinLock(&Adapter->IntLock);

	MP_DEC_REF(Adapter);	 

	SMSC_TRACE0(DBG_INT,"<-- NICPostIntsWorkItemCallBack\r\n"); 	   
}


/*++

Routine Description:

	This function submits all interrupt requests.
	
Arguments:

	Adapter    - pointer to the MP_ADAPTER structure
	
Return value:

	NIC_STATUS_FAILURE if reset pipe failed, NIC_STATUS_SUCCESS otherwise
--*/

NIC_STATUS
NICStartInterruptPipe(
	PMP_ADAPTER Adapter
	)
{ 
	NIC_STATUS		nStatus = NIC_STATUS_SUCCESS;

	SMSC_TRACE0(DBG_INT,"--> NICStartInterruptPipe\r\n");
	

	// We need to reset the pipe to make sure the data toggles are in sync.
	if(!NICResetPipe(Adapter, Adapter->Interrupt)) 
	{
		SMSC_TRACE0(DBG_ERROR, "Interrupt Pipe Reset failed !!! Aborting Int Pipe Start sequence.\r\n");
		nStatus = NIC_STATUS_FAILURE;
		goto Exit_NICStartInterruptPipe;
	}

	Adapter->Interrupt.PipeTransferAborted=0;
	NdisSetEvent(&Adapter->IsIntsCompletionWorkItemReady);

	MP_SET_FLAG(Adapter, fMP_POST_INTS);

	Adapter->IsIntReady=TRUE;
	NICPostInts(Adapter);

Exit_NICStartInterruptPipe:
	SMSC_TRACE0(DBG_INT,"<-- NICStartInterruptPipe\r\n");

	return (nStatus);
}


/*++
Routine Description:

   Post all the free int requests on USB if the flag to do so is enabled.
	  
Arguments:

	Adapter    - pointer to the MP_ADAPTER structure
	
Return Value:

	None.

--*/
void
NICPostInts(
	PMP_ADAPTER Adapter
	)

{
	
	SMSC_TRACE0(DBG_INT,"--> NICPostInts\r\n");

	// If a workitem already scheduled don't queue another.		
	if (InterlockedExchange(&Adapter->IsPostIntsWorkItemQueued, TRUE) == FALSE)
	{
		Adapter->nPostIntsWorkItemScheduled++;
		MP_INC_REF(Adapter);				   
		NdisScheduleWorkItem(&Adapter->PostIntsWorkItem);	
	}

	SMSC_TRACE0(DBG_INT,"<-- NICPostInts\r\n");
}


/*++

Routine Description:

	This function stops further submissions on the interrupt pipe, 
	cancels all pending URBs and waits for their completion.
	
Arguments:

	Adapter    - pointer to the MP_ADAPTER structure
	
Return value:

	VOID
--*/

NIC_STATUS
NICStopInterruptPipe(
	PMP_ADAPTER Adapter
	)
{ 
	BOOLEAN 		  bDone=TRUE;
	LONG			  nWaitCount = 0;
	NIC_STATUS		  Status=NIC_STATUS_SUCCESS;

	SMSC_TRACE0(DBG_SHUT_DOWN,"--> NICStopInterruptPipe\r\n");

	if (Adapter->IsIntReady)
	 	NdisWaitEvent(&Adapter->IsIntsCompletionWorkItemReady,0);

	// Since it can be called from MP_Halt when MP_Initialize fails
	if(!MP_TEST_FLAG(Adapter, fMP_INT_RESOURCE_ALLOCATED)){
		Status=NIC_STATUS_FAILURE;
		goto Exit_NICStopInterruptPipe;
	}

	MP_CLEAR_FLAG(Adapter, fMP_POST_INTS);
	
	if (!MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED)) {
		// Try to stop them before the completion in one shot.
		Adapter->Interrupt.PipeTransferAborted=TRUE;
		if ( ! Adapter->UsbFuncs->lpAbortPipeTransfers(Adapter->Interrupt.hPipe,USB_NO_WAIT))
		{
			SMSC_TRACE0(DBG_ERROR,"Interrupt Pipe Abort failed !!!\r\n");		
		}
	}
	
	// Wait for all int completions and ICBs to make it back to the free list.
	while(TRUE)
	{
		bDone = TRUE;
	   
		
		NdisAcquireSpinLock(&Adapter->IntLock);

		//
		// Are all the interrupt trancastions completed ?
		//		
		if(Adapter->nBusyInt + Adapter->nProcessedInt + Adapter->nCompletionInt)
		{
			SMSC_TRACE1(DBG_INT,"nBusyInt = %d\r\n", Adapter->nBusyInt);
			SMSC_TRACE1(DBG_INT,"nProcessedInt = %d\r\n", Adapter->nProcessedInt);
			SMSC_TRACE1(DBG_INT,"nCompletionInt = %d\r\n", Adapter->nCompletionInt);
			bDone = FALSE;
		}
		NdisReleaseSpinLock(&Adapter->IntLock); 

		if(bDone)
		{
			break;	 
		}

		if((++nWaitCount % NIC_1_SEC_TIMEOUT) == 0)
		{
			SMSC_TRACE1(DBG_ERROR,"Int Cancel timed out after %d second!!! COMMITING LEAK !!!\r\n", NIC_1_SEC_TIMEOUT/1000);
			SMSC_TRACE1(DBG_ERROR, "   Adapter->nBusyInt = %d\r\n", Adapter->nBusyInt);
			SMSC_TRACE1(DBG_ERROR, "   Adapter->nProcessedInt = %d\r\n", Adapter->nProcessedInt);
			SMSC_TRACE1(DBG_ERROR,"Adapter->nCompletionInt = %d\r\n", Adapter->nCompletionInt);
			Status=NIC_STATUS_FAILURE;
			ASSERT(FALSE);		 
			break;	
		}

		NdisMSleep(NDIS_SLEEP_1_MS);		
	}

	
Exit_NICStopInterruptPipe:
	
	SMSC_TRACE0(DBG_SHUT_DOWN,"<-- NICStopInterruptPipe\r\n");
	return Status;
}

/*++

Routine Description:

   Workitem to service interrupts. This workitme is scheduled in the 
   interrupt pipe completion routine each time we receive an interrupt 
   with status other than "receive data in receive fifo" 
	  
Arguments:

	WorkItem - Pointer to workitem
	
	Context - Pointer to our Adapter

Return Value:

	None.

--*/
VOID
NICIntStatWorkItemCallBack(
	PNDIS_WORK_ITEM WorkItem, 
	PVOID Context
	)
{
	PMP_ADAPTER 	Adapter = (PMP_ADAPTER)Context;
	DWORD			Value32;
	PICB			pICB=NULL;
		   
	SMSC_TRACE0(DBG_INT,"--> NICIntWorkItemCallBack\r\n");

	// Read the interrupt Status again now and process the sources.
	if (NICReadRegister(Adapter, ETH_INT_STS, &Value32) != NIC_STATUS_SUCCESS)
		goto Exit_NICIntStatWorkItemCallBack;
	
	if (Value32 & (~(ETH_INT_STS_RXDF_ | ETH_INT_STS_GPIOS_)))
	{
		SMSC_TRACE1(DBG_ERROR, "Interrupt Status = 0x%08lx\r\n", Value32); 

		if (Value32 & ETH_INT_STS_MACRTO_)
		{
			SMSC_TRACE0(DBG_ERROR, "   ETH_INT_STS_MACRTO_: MAC Reset Timeout, Should never happen !\r\n"); 
		}
		if (Value32 & ETH_INT_STS_TX_STOP_)
		{
			SMSC_TRACE0(DBG_ERROR,"   ETH_INT_STS_TX_STOP_: Tx Stopped\r\n");
		}
		if (Value32 & ETH_INT_STS_RX_STOP_)
		{
			SMSC_TRACE0(DBG_ERROR,"   ETH_INT_STS_RX_STOP_: Rx Stopped\r\n");
		}
		if (Value32 & ETH_INT_STS_PHY_INT_)
		{
			SMSC_TRACE0(DBG_INT,"	ETH_INT_STS_PHY_INT_: PHY int\r\n");
		}
		if (Value32 & ETH_INT_STS_TXE_)
		{
			SMSC_TRACE0(DBG_ERROR,"   ETH_INT_STS_TXE_	: TX Error, will also cause pipe stall. Handled by USB Error recovery.\r\n");
		}
		if (Value32 & ETH_INT_STS_TDFU_)
		{
			SMSC_TRACE0(DBG_ERROR,"   ETH_INT_STS_TDFU_: TX Underrun, Should never happen !\r\n");
		}
		if (Value32 & ETH_INT_STS_TDFO_	)
		{
			SMSC_TRACE0(DBG_ERROR,"   ETH_INT_STS_TDFO_: TX Overrun, Should never happen !\r\n");
		}
		if (Value32 & ETH_INT_STS_RXDF_	)
		{
			SMSC_TRACE0(DBG_ERROR,"   ETH_INT_STS_RXDF_: Rx Dropped frame. \r\n");
		}
	}

	NICWriteRegister(Adapter, ETH_INT_STS, &Value32);
	
Exit_NICIntStatWorkItemCallBack:

	//
	// Clear the flag to let the WorkItem structure be reused for
	// scheduling another one.
	//
	InterlockedExchange(&Adapter->IsIntStatWorkItemQueued, FALSE);

	MP_DEC_REF(Adapter);	 
	
	SMSC_TRACE0(DBG_INT,"<-- NICIntWorkItemCallBack\r\n");		  
}

/*++

Routine Description:

	This routine sends a interrupt request to the USB
		
Arguments:

	Adapter    - pointer to the MP_ADAPTER structure
	pICB	-  Pointer to the ICB block 


Return Value:

	NT status code

--*/
NTSTATUS
NICPostIntRequest(
	PMP_ADAPTER Adapter,
	PICB	pICB
	)
{

//	PDEVICE_OBJECT	 TargetDeviceObject = Adapter->TargetDeviceObject;
	int count=0;
			  
	SMSC_TRACE1(DBG_INT,"--> NICPostIntRequest (pICB = %p)\r\n", pICB);

	if (!ACCEPT_IO(Adapter)) {
		pICB->Ref = 1;
		SMSC_TRACE0(DBG_ERROR,"NICPostIntRequest(),Device is not ready for I/O request!!!\r\n");
		return STATUS_UNSUCCESSFUL;
	}
		
	ResetEvent(pICB->SyncEvent);
	pICB->Ref = 1;
	pICB->dwErr= IssueInterruptTransfer( Adapter->UsbFuncs,
									   Adapter->Interrupt.hPipe,
									   NICIntRequestCompletion,
									   pICB,
									   (USB_IN_TRANSFER|USB_SHORT_TRANSFER_OK),
									   pICB->pData, 0,
									   INT_BUFFER_SIZE,
									  &pICB->dwTransferLength,
									   0,
									  &pICB->dwUsbErr);
			
	 SetEvent(pICB->SyncEvent);

//	 ASSERT(!(!pICB->dwErr && !pICB->dwUsbErr));
	 if ( (ERROR_SUCCESS ==  pICB->dwErr) && (USB_NO_ERROR ==  pICB->dwUsbErr)	) 
	 {

		SMSC_TRACE2(DBG_ERROR,"Interrupt Request, NULL Transfer,  pICB->dwErr:%d  pICB->dwUsbErr:0x%x)\r\n",  pICB->dwErr,  pICB->dwUsbErr);
		return STATUS_UNSUCCESSFUL;
			   
	 }

		
	 if ((ERROR_SUCCESS != pICB->dwErr)) {
		SMSC_TRACE2(DBG_ERROR,"IssueInterruptTransfer ERROR, dwErr= 0x%x, dwUsbErr=0x%x\r\n",pICB->dwErr,pICB->dwUsbErr);
		return STATUS_UNSUCCESSFUL;
	 }
	
	
	SMSC_TRACE0(DBG_INT,"<-- NICPostIntRequest\r\n");

	return STATUS_SUCCESS;
}


/*++

Routine Description:

	Completion routine for the Int request. This routine
	put the ICB into IntcompletionList and open another workitem to process it
		
Arguments:

	pICB -	pointer to the context structure (ICB) 

Return Value:

	STATUS_SUCCESS
	
--*/
DWORD
NICIntRequestCompletion(
	PICB   pICB
	)
{

	PMP_ADAPTER Adapter = pICB->Adapter;
	
	SMSC_TRACE0(DBG_INT,"--> NICIntRequestCompletion \r\n");

	NdisAcquireSpinLock(&Adapter->IntLock);

	// Remove it from the busy list 
	// and put it in the completion List
	RemoveEntryList(&pICB->List);	
	NdisInterlockedDecrement(&Adapter->nBusyInt);
	ASSERT(Adapter->nBusyInt >= 0);

	// Put it in the completion list.
	InsertTailList(&Adapter->IntCompletionList, &pICB->List);
	NdisInterlockedIncrement(&Adapter->nCompletionInt);
	ASSERT(Adapter->nCompletionInt <= Adapter->MaxICBs);

	SMSC_TRACE1(DBG_INT,"ICB %p removed from it's busy list and inserted into IntCompletionList\r\n", pICB);

	NdisReleaseSpinLock(&Adapter->IntLock);

	#ifdef USE_THREAD
		SetEvent(Adapter->IntThreadEvent);
	#else
		#ifdef USE_WORK_ITEM_NO_OUT
		 NdisSetEvent(&Adapter->IntsWorkItemReady);
		#else
		
		if (InterlockedExchange(&Adapter->IsIntsCompletionWorkItemQueued, TRUE) == FALSE)
		{
			Adapter->nIntsCompletionWorkItemScheduled++;
			MP_INC_REF(Adapter);
			NdisScheduleWorkItem(&Adapter->IntsCompletionWorkItem);
		}
		#endif
	#endif
	
	SMSC_TRACE0(DBG_INT,"<-- NICIntRequestCompletion\r\n");

	return STATUS_SUCCESS;
}

/*++

Routine Description:

	This routine process the ICBs in the IntCompletionList 
		
Arguments:

	Context  point to the adapter
Return Value:

	STATUS_SUCCESS
	
--*/

NTSTATUS
IntsCompletionWorkItemCallBack(
   PNDIS_WORK_ITEM WorkItem, 
	PVOID Context
	)
{
	PMP_ADAPTER Adapter = Context;
	PICB pICB;
	BOOLEAN	bIntFailed = FALSE;
	ULONG	bytesRead;
	ULONG	dwInt;
	USB_ERROR	UsbErr;
	USB_TRANSFER hTransfer;
	BOOL dwErr=FALSE;
	
	SMSC_TRACE0(DBG_INT,"--> IntsCompletionWorkItemCallback. \r\n");
	
	NdisWaitEvent(&Adapter->IsIntsCompletionWorkItemReady, 0);
	NdisResetEvent(&Adapter->IsIntsCompletionWorkItemReady);

	InterlockedExchange(&Adapter->IsIntsCompletionWorkItemQueued, FALSE);


	while (!IsListEmpty(&Adapter->IntCompletionList)) {
	
		NdisAcquireSpinLock(&Adapter->IntLock);
		pICB = (PICB) RemoveHeadList(&Adapter->IntCompletionList);
		ASSERT(pICB && (pICB != (PICB) &Adapter->IntCompletionList));// cannot be NULL and list cannot be empty
		NdisInterlockedDecrement(&Adapter->nCompletionInt);
		ASSERT(Adapter->nCompletionInt >= 0);
		NdisReleaseSpinLock(&Adapter->IntLock);

		//Sometimes the completion routine is called before the transfer complete. We need to sync. it here.
		//wait for the IssueInterruptTransfer() to complete before we process the interrupt transfer	handler
		WaitForSingleObject(pICB->SyncEvent, 100);
		ResetEvent(pICB->SyncEvent);

	
		if (pICB->dwUsbErr == 0) {
			SMSC_TRACE1(DBG_ERROR,"Interrupt Request has NULL Transfer,  hTransfer=0x%x \r\n", pICB->dwUsbErr);	
		}


		hTransfer=(USB_TRANSFER)pICB->dwUsbErr;
		dwErr=GetTransferStatus(Adapter->UsbFuncs, hTransfer, &pICB->dwTransferLength,&UsbErr);


		if (hTransfer) {
			CloseTransferHandle(Adapter->UsbFuncs, hTransfer);
		}

		bytesRead = pICB->dwTransferLength;
	
		if( !(USB_NO_ERROR == UsbErr) || !dwErr || Adapter->Interrupt.PipeTransferAborted)
		{
			SMSC_TRACE3(DBG_ERROR, "Int request failed UsbErr = %x, dwErr = %x, hTransfer=IF0x%x\r\n", 
					UsbErr,
					dwErr,
					hTransfer);
						
			bIntFailed = TRUE;
		}
		else if (bytesRead != 4)
		{
			SMSC_TRACE3(DBG_ERROR, "Int request invalid size = %d , Data Buffer: 0x%p, data= 0x%x\r\n", bytesRead, pICB->pData, *(pICB->pData));		
			bIntFailed = TRUE;
		}
	

		if (bIntFailed)
		{

			Adapter->nIntsCompletedWithError++; 	   

			//
			// Clear the flag to prevent any more ints from being
			// posted to the target device.
			//
			MP_CLEAR_FLAG(Adapter, fMP_POST_INTS);

			// Handle USB Error recovery
			if ((UsbErr != USB_CANCELED_ERROR) && \
				(!Adapter->Interrupt.PipeTransferAborted) &&   \
				(!MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS)) && \
				(!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS)) && \
				(!MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED)) && \
				(!MP_TEST_FLAG(Adapter, fMP_DEVICE_NOT_ATTACHED)) )
			{

				SMSC_TRACE3(DBG_ERROR, "Int request failed UsbErr = %x, dwErr = %x, hTransfer=IF0x%x\r\n", 
										UsbErr,
										dwErr,
										hTransfer);
			
				if (InterlockedExchange(&Adapter->IsUsbErrorRecoveryWorkItemQueued, TRUE) == FALSE) 
				{
						Adapter->nUsbErrorRecoveryWorkItemScheduled++;
						MP_INC_REF(Adapter);				   
					NdisScheduleWorkItem(&Adapter->UsbErrorRecoveryWorkItem);
				}
			}		
	   
		
		}
		else
		{
			// We got a good Interrupt transaction completion, count it.
			Adapter->nIntsCompletedSuccessfully++;
			dwInt = *(DWORD*) pICB->pData;
			SMSC_TRACE2(DBG_INT,"Int %d bytes = 0x%08lx\r\n", bytesRead, dwInt);
		}
 
		// Let's see if we should actually process it.
		if (MP_TEST_FLAG(Adapter, fMP_POST_INTS))	// Checking for this is enough since only NICStartInterruptPipe sets it
		{
			if (!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS) && !MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS))
			{
				if ((Adapter->StopBulkInOnRxEmpty) && (dwInt & INT_ENP_RFHF_)) 
				{
				MP_SET_FLAG(Adapter, fMP_POST_READS);
	
				// For performance, instead of scheduling reads at the end of
				// every int completion that showed data is available, we will 
				// do it only when the number of outstanding ICBs goes below 
				// Adapter->LowWatermarkICBs.
//				if((!Adapter->LowWatermarkICBs) || (Adapter->nBusyRecv <= Adapter->LowWatermarkICBs))
				{
					NICPostReads(Adapter);
				}
			}

			// We'll schedule an Int Status workitem if the interrupt itself
			// needs processing
			if (dwInt & ~(INT_ENP_RFHF_ | ETH_INT_STS_GPIOS_))
			{
				// If a workitem already scheduled don't queue another.
				if (InterlockedExchange(&Adapter->IsIntStatWorkItemQueued, TRUE) == FALSE)
				{
					Adapter->nIntStatWorkItemScheduled++;
					MP_INC_REF(Adapter);				   
					NdisScheduleWorkItem(&Adapter->IntStatWorkItem);
				}
			}
		}
	}

	// We're done with this ICB
	NdisAcquireSpinLock(&Adapter->IntLock);

	// Put it in the Processed list.
	InsertTailList(&Adapter->IntProcessedList, &pICB->List);
	NdisInterlockedIncrement(&Adapter->nProcessedInt);
	ASSERT(Adapter->nProcessedInt <= Adapter->MaxICBs);

	SMSC_TRACE1(DBG_INT,"ICB %p inserted into IntProcessedList\r\n", pICB);

	NdisReleaseSpinLock(&Adapter->IntLock); 

	if (NdisInterlockedDecrement(&pICB->Ref) == 0) 
	{
		NICFreeICB(pICB);
	}
	else
	{
		// This should never happen
		SMSC_TRACE0(DBG_ERROR, "NICIntRequestCompletion finds ICB with invalid Ref count!\r\n");			 
	}
}
	
	MP_DEC_REF(Adapter); 
	NdisSetEvent(&Adapter->IsIntsCompletionWorkItemReady);
	SMSC_TRACE0(DBG_INT,"<-- IntCompletionWorkItemCallback\r\n");

	return STATUS_SUCCESS;
}



/*++

Routine Description:

	This routine process the ICBs in the IntCompletionList 
		
Arguments:

	Context  point to the adapter
Return Value:

	STATUS_SUCCESS
	
--*/

NTSTATUS
IntsCompletionCallBack(
   PMP_ADAPTER Adapter
	)
{

	PICB pICB;
	BOOLEAN	bIntFailed = FALSE;
	ULONG	bytesRead;
	ULONG	dwInt;
	USB_ERROR	UsbErr;
	USB_TRANSFER hTransfer;
	BOOL dwErr=FALSE;
	
	SMSC_TRACE0(DBG_INT,"--> IntsCompletionWorkItemCallback. \r\n");
	
	if (Adapter->MPInitializeFail ||(!MP_TEST_FLAG(Adapter, fMP_INT_RESOURCE_ALLOCATED)))
				return STATUS_SUCCESS;


	while (!IsListEmpty(&Adapter->IntCompletionList)) {
	
		NdisAcquireSpinLock(&Adapter->IntLock);
		pICB = (PICB) RemoveHeadList(&Adapter->IntCompletionList);
		ASSERT(pICB && (pICB != (PICB) &Adapter->IntCompletionList));// cannot be NULL and list cannot be empty
		NdisInterlockedDecrement(&Adapter->nCompletionInt);
		ASSERT(Adapter->nCompletionInt >= 0);
		NdisReleaseSpinLock(&Adapter->IntLock);

		//Sometimes the completion routine is called before the transfer complete. We need to sync. it here.
		//wait for the IssueInterruptTransfer() to complete before we process the interrupt transfer	handler
		WaitForSingleObject(pICB->SyncEvent, 100);
		ResetEvent(pICB->SyncEvent);

	
		if (pICB->dwUsbErr == 0) {
			SMSC_TRACE1(DBG_ERROR,"Interrupt Request has NULL Transfer,  hTransfer=0x%x \r\n", pICB->dwUsbErr);	
		}


		hTransfer=(USB_TRANSFER)pICB->dwUsbErr;
		dwErr=GetTransferStatus(Adapter->UsbFuncs, hTransfer, &pICB->dwTransferLength,&UsbErr);


		if (hTransfer) {
			CloseTransferHandle(Adapter->UsbFuncs, hTransfer);
		}

		bytesRead = pICB->dwTransferLength;
	
		if( !(USB_NO_ERROR == UsbErr) || !dwErr || Adapter->Interrupt.PipeTransferAborted)
		{
			SMSC_TRACE3(DBG_ERROR, "Int request failed UsbErr = %x, dwErr = %x, hTransfer=IF0x%x\r\n", 
					UsbErr,
					dwErr,
					hTransfer);
						
			bIntFailed = TRUE;
		}
		else if (bytesRead != 4)
		{
			SMSC_TRACE3(DBG_ERROR, "Int request invalid size = %d , Data Buffer: 0x%p, data= 0x%x\r\n", bytesRead, pICB->pData, *(pICB->pData));		
			bIntFailed = TRUE;
		}
	

		if (bIntFailed)
		{

			Adapter->nIntsCompletedWithError++; 	   

			//
			// Clear the flag to prevent any more ints from being
			// posted to the target device.
			//
			MP_CLEAR_FLAG(Adapter, fMP_POST_INTS);

			// Handle USB Error recovery
			if ((UsbErr != USB_CANCELED_ERROR) && \
				(!Adapter->Interrupt.PipeTransferAborted) &&   \
				(!MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS)) && \
				(!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS)) && \
				(!MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED)) && \
				(!MP_TEST_FLAG(Adapter, fMP_DEVICE_NOT_ATTACHED)) )
			{

				SMSC_TRACE3(DBG_ERROR, "Int request failed UsbErr = %x, dwErr = %x, hTransfer=IF0x%x\r\n", 
										UsbErr,
										dwErr,
										hTransfer);
			
				if (InterlockedExchange(&Adapter->IsUsbErrorRecoveryWorkItemQueued, TRUE) == FALSE) 
				{
						Adapter->nUsbErrorRecoveryWorkItemScheduled++;
						MP_INC_REF(Adapter);				   
					NdisScheduleWorkItem(&Adapter->UsbErrorRecoveryWorkItem);
				}
			}		
	   
		
		}
		else
		{
			// We got a good Interrupt transaction completion, count it.
			Adapter->nIntsCompletedSuccessfully++;
			dwInt = *(DWORD*) pICB->pData;
			SMSC_TRACE2(DBG_INT,"Int %d bytes = 0x%08lx\r\n", bytesRead, dwInt);
		}
 
		// Let's see if we should actually process it.
		if (MP_TEST_FLAG(Adapter, fMP_POST_INTS))	// Checking for this is enough since only NICStartInterruptPipe sets it
		{
			if (!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS) && !MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS))
			{
				if ((Adapter->StopBulkInOnRxEmpty) && (dwInt & INT_ENP_RFHF_)) 
				{
				MP_SET_FLAG(Adapter, fMP_POST_READS);
	
				// For performance, instead of scheduling reads at the end of
				// every int completion that showed data is available, we will 
				// do it only when the number of outstanding ICBs goes below 
				// Adapter->LowWatermarkICBs.
//				if((!Adapter->LowWatermarkICBs) || (Adapter->nBusyRecv <= Adapter->LowWatermarkICBs))
				{
					NICPostReads(Adapter);
				}
			}

			// We'll schedule an Int Status workitem if the interrupt itself
			// needs processing
			if (dwInt & ~(INT_ENP_RFHF_ | ETH_INT_STS_GPIOS_))
			{
				// If a workitem already scheduled don't queue another.
				if (InterlockedExchange(&Adapter->IsIntStatWorkItemQueued, TRUE) == FALSE)
				{
					Adapter->nIntStatWorkItemScheduled++;
					MP_INC_REF(Adapter);				   
					NdisScheduleWorkItem(&Adapter->IntStatWorkItem);
				}
			}
		}
	}

	// We're done with this ICB
	NdisAcquireSpinLock(&Adapter->IntLock);

	// Put it in the Processed list.
	InsertTailList(&Adapter->IntProcessedList, &pICB->List);
	NdisInterlockedIncrement(&Adapter->nProcessedInt);
	ASSERT(Adapter->nProcessedInt <= Adapter->MaxICBs);

	SMSC_TRACE1(DBG_INT,"ICB %p inserted into IntProcessedList\r\n", pICB);

	NdisReleaseSpinLock(&Adapter->IntLock); 

	if (NdisInterlockedDecrement(&pICB->Ref) == 0) 
	{
		NICFreeICB(pICB);
	}
	else
	{
		// This should never happen
		SMSC_TRACE0(DBG_ERROR, "NICIntRequestCompletion finds ICB with invalid Ref count!\r\n");			 
	}
}
	
	SMSC_TRACE0(DBG_INT,"<-- IntCompletionWorkItemCallback\r\n");

	return STATUS_SUCCESS;
}


/*++

Routine Description:

	This routine reinitializes the ICB block and puts it back
	into the IntFreeList for reuse. Assumes the ICB was in the 
	process list
		
Arguments:

	pICB	  - pointer to ICB block   

Return Value:

	VOID

--*/
VOID 
NICFreeICB(
	IN PICB pICB
	)
{
	PMP_ADAPTER Adapter = pICB->Adapter;

	SMSC_TRACE1(DBG_INT,"--> NICFreeICB %p\r\n", pICB);
	
	ASSERT(!pICB->Ref); // should be 0
	ASSERT(pICB->Adapter); // shouldn't be NULL

	NdisAcquireSpinLock(&Adapter->IntLock);
	
	RemoveEntryList(&pICB->List);
	NdisInterlockedDecrement(&Adapter->nProcessedInt);
	ASSERT(Adapter->nProcessedInt >= 0);
   
	SMSC_TRACE1(DBG_INT,"ICB %p removed from the Processed List and inserted into IntFreeList\r\n", pICB);
	InsertTailList(&Adapter->IntFreeList, &pICB->List);

	NdisReleaseSpinLock(&Adapter->IntLock); 

	if (!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS) && !MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS) && !MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED))
	{
		// For performance, unless the caller specifically requested 
		// a workitem, instead of scheduling reads at the end of
		// every read completion, we will do it only when the number of 
		// outstanding ICBs goes below Adapter->LowWatermarkICBs.
		if(
			((!Adapter->LowWatermarkICBs) || (Adapter->nBusyInt <= Adapter->LowWatermarkICBs))
			&& MP_TEST_FLAG(Adapter, fMP_POST_INTS)
		  )
		{
			NICPostInts(Adapter);
		}
	}

	SMSC_TRACE1(DBG_INT,"<-- NICFreeICB %d\r\n", Adapter->nBusyInt);
}

#endif

/*++

Routine Description:
	
	This is the Linkcheck Timer Callback function. 

Arguments:

	FunctionContext - Pointer to our Adapter
	Others			- Not used 

Return Value:

	None.

--*/

VOID 
NICLinkCheck(
	IN PVOID  SystemSpecific1,
	IN PVOID  FunctionContext,
	IN PVOID  SystemSpecific2,
	IN PVOID  SystemSpecific3
	)
{
	PMP_ADAPTER Adapter = (PMP_ADAPTER) FunctionContext;

	UNREFERENCED_PARAMETER(SystemSpecific1);
	UNREFERENCED_PARAMETER(SystemSpecific2);
	UNREFERENCED_PARAMETER(SystemSpecific3);
	NdisAcquireSpinLock(&Adapter->LCSpinlock);

	SMSC_TRACE0(DBG_LINK,"--> NICLinkCheck\r\n");

	if (MP_IS_READY(Adapter))
	{
		
		if (InterlockedExchange(&Adapter->IsLinkCheckWorkItemQueued, TRUE) == FALSE)
		{
			MP_SET_FLAG(Adapter, fMP_LINK_CHECK_IN_PROGRESS);
			Adapter->nLinkCheckWorkItemScheduled++;
			SMSC_TRACE0(DBG_LINK,"	 Callback Scheduled\r\n");
			MP_INC_REF(Adapter);				   
			NdisScheduleWorkItem(&Adapter->LinkCheckWorkItem);	 
		}
		
		else
		{
			SMSC_TRACE0(DBG_LINK,"	 NOT Scheduling callback, one still pending.\r\n");
		}
		
	}

	// Signal any thread waiting for the Timer Cancellation that we're done.
	NdisSetEvent(&Adapter->LinkCheckCalledEvent);

	NdisReleaseSpinLock(&Adapter->LCSpinlock);

	SMSC_TRACE0(DBG_LINK,"<-- NICLinkCheck\r\n");
}

int Nic_phy_boost(PMP_ADAPTER Adapter, int mode)
{
	DWORD dwTemp;
	int Count;

	if ((Adapter->PhyHookup == PHY_HOOKUP_INTERNAL) &&
		(GetChipIdFromID_REV(Adapter->DevIdReg) == ETH_ID_REV_CHIP_ID_9730_) || 
		(GetChipIdFromID_REV(Adapter->DevIdReg) == ETH_ID_REV_CHIP_ID_9530_) || 
		(GetChipIdFromID_REV(Adapter->DevIdReg) == ETH_ID_REV_CHIP_ID_9E08_)) 
	{
		if (mode == 1) {
            SMSC_TRACE0(DBG_INIT,"Setting phy boost\r\n");
			/* set boost */
			dwTemp = PHY_THERA_TSTCNTL_TEST_MODE;
			NICWritePhy(Adapter, PHY_THERA_TSTCNTL, dwTemp);	
			NICWritePhy(Adapter, PHY_THERA_TSTCNTL, 0 );	
			NICWritePhy(Adapter, PHY_THERA_TSTCNTL, dwTemp);	

			dwTemp = 0x85E8;
			NICWritePhy(Adapter, PHY_THERA_TSTWRITE, dwTemp);

            for(Count=1;Count<10;Count++){
				dwTemp = 0;
				NICReadPhy(Adapter, PHY_THERA_TSTCNTL, &dwTemp);	
				if(!(dwTemp & PHY_THERA_TSTCNTL_WRITE))
					break;
				NdisMSleep(NDIS_SLEEP_10_MS);
			}

			if (Count >= 10){
				SMSC_TRACE0(DBG_LINK,"Timed out writing PHY_TSTWRITE\r\n");
			}
			//dwTemp = 0x4416;
			dwTemp = (PHY_THERA_TSTCNTL_WRITE | PHY_THERA_TSTCNTL_TEST_MODE | 0x16);
			NICWritePhy(Adapter, PHY_THERA_TSTCNTL, dwTemp);	
			/* Loop until write is completed w/timeout */
			for(Count=1;Count<10;Count++){
				dwTemp = 0;
				NICReadPhy(Adapter, PHY_THERA_TSTCNTL, &dwTemp);	
				if(!(dwTemp & PHY_THERA_TSTCNTL_WRITE))
					break;
				NdisMSleep(NDIS_SLEEP_10_MS);
			}

			if (Count >= 10){
				SMSC_TRACE0(DBG_LINK,"Timed out writing PHY_TSTWRITE\r\n");
			}

			#if 0 /* for DEBUG */
			/* read and verify the value */
			dwTemp = 0x86C0;
			NICWritePhy(Adapter, PHY_THERA_TSTCNTL, dwTemp);	
			/* Loop until read is completed w/timeout */
			for(Count=1;Count<100;Count++){
				dwTemp = 0;
				NICReadPhy(Adapter, PHY_THERA_TSTCNTL, &dwTemp);	
				if(!(dwTemp & PHY_THERA_TSTCNTL_READ))
					break;
				NdisMSleep(NDIS_SLEEP_10_MS);
			}

			/* if bXXXXXX11XXXXXXXX, then VOH Compliance is enabled. While doing experiment
			   some time the bit 8 was not set. But still found that phy boost is working.
			   commenting out this checking.
			  */
			if (Count < 100){
				dwTemp = 0;
				NICReadPhy(Adapter, PHY_THERA_TSTREAD1, &dwTemp);	
				if ((dwTemp & 0x300) != 0x300){
					SMSC_TRACE1(DBG_LINK,"Setting Ethernet Boost Successful: 0x%x \r\n", dwTemp);
				} else{
					SMSC_TRACE1(DBG_LINK,"Setting Ethernet Boost Failed: 0x%x \r\n", dwTemp);
				}
			
			}else{
				SMSC_TRACE0(DBG_LINK,"Timed out reading PHY_TSTREAD1 register\r\n");
				
			}
			#endif

            NICWritePhy(Adapter, PHY_THERA_TSTCNTL, 0 );
				
		} else {
			#if 0 /* for DEBUG */
            dwTemp = PHY_THERA_TSTCNTL_TEST_MODE;
			NICWritePhy(Adapter, PHY_THERA_TSTCNTL, dwTemp);	
			NICWritePhy(Adapter, PHY_THERA_TSTCNTL, 0 );	
			NICWritePhy(Adapter, PHY_THERA_TSTCNTL, dwTemp);

			/* read and verify the value */
			dwTemp = 0x86C0;
			NICWritePhy(Adapter, PHY_THERA_TSTCNTL, dwTemp);	
			/* Loop until read is completed w/timeout */
			for(Count=1;Count<100;Count++){
				dwTemp = 0;
				NICReadPhy(Adapter, PHY_THERA_TSTCNTL, &dwTemp);	
				if(!(dwTemp & PHY_THERA_TSTCNTL_READ))
					break;
				NdisMSleep(NDIS_SLEEP_10_MS);
			}
			/* if bXXXXXX11XXXXXXXX, then VOH Compliance is enabled. While doing experiment
			   some time the bit 8 was not set. But still found that phy boost is working.
			   commenting out this part of the code
			   */
			if (Count < 100){
				dwTemp = 0;
				NICReadPhy(Adapter, PHY_THERA_TSTREAD1, &dwTemp);	
				if ((dwTemp & 0x300) != 0x300){
					SMSC_TRACE1(DBG_LINK,"Ethernet Boost: 0x%x \r\n", dwTemp);
				} else{
					SMSC_TRACE1(DBG_LINK,"Failed to set Ethernet Boost: 0x%x \r\n", dwTemp);
				}
			
			}else{
				SMSC_TRACE0(DBG_LINK,"Timed out reading PHY_TSTREAD1 register\r\n");
				
			}

            NICWritePhy(Adapter, PHY_THERA_TSTCNTL, 0 );
			#endif
        }
	}
	return 0;
}

__inline
VOID
NICUpdateStatCounters(
    IN PMP_ADAPTER Adapter,
    IN PUCHAR pStats
    )
/*++

Routine Description:

    This routine updates Tx/Rx statistic counters.

Arguments:

    Adapter - pointer to our Adapter
    pStats - Pointer to stat. counters array

Return Value:

--*/
{
    PLAN9500_RX_STATS   pRxStats = (PLAN9500_RX_STATS) pStats;
    PLAN9500_TX_STATS   pTxStats = (PLAN9500_TX_STATS) (pStats + sizeof(LAN9500_RX_STATS));
    BOOLEAN fRolloverCounters = FALSE;
    DWORD Temp32;

    if (NIC_STATUS_SUCCESS != NICReadRegister(Adapter, ETH_ID_REV, &Temp32))
    {
        return;
    }

    switch (GetChipIdFromID_REV(Temp32))
    {
        // Clear-On-Read Stat. Counters
        case ETH_ID_REV_CHIP_ID_9500_:
        case ETH_ID_REV_CHIP_ID_9505_:
        case ETH_ID_REV_CHIP_ID_EC00_:
            break;

        // Rollover Stat. Counters
        case ETH_ID_REV_CHIP_ID_9E00_:
        case ETH_ID_REV_CHIP_ID_9E01_:
        case ETH_ID_REV_CHIP_ID_9E08_:
        case ETH_ID_REV_CHIP_ID_9E09_:
        case ETH_ID_REV_CHIP_ID_9730_:
        case ETH_ID_REV_CHIP_ID_9735_:
        case ETH_ID_REV_CHIP_ID_9530_:
        case ETH_ID_REV_CHIP_ID_9535_:
        default: // --fixme::Assuming rollover counters for Device IDs added in registry
        
            fRolloverCounters = TRUE;
            break;
    }
    
    if (fRolloverCounters == TRUE)
    {
        // Good Tx (32Bit Counter)
        if (pTxStats->TxGoodFrames >= Adapter->TxStats.TxGoodFrames)
            Adapter->TransmitsGood += (ULONG64)(pTxStats->TxGoodFrames - Adapter->TxStats.TxGoodFrames);
        else
            Adapter->TransmitsGood += (ULONG64)(pTxStats->TxGoodFrames + (0xFFFFFFFFUL - Adapter->TxStats.TxGoodFrames) + 1UL);

        Adapter->TxStats.TxGoodFrames = pTxStats->TxGoodFrames;

        // Tx Excess Collisions (20Bit Counter)
        if (pTxStats->TxExcessiveCollisionErrors >= Adapter->TxStats.TxExcessiveCollisionErrors)
            Adapter->TxAbortExcessCollisions += (pTxStats->TxExcessiveCollisionErrors - Adapter->TxStats.TxExcessiveCollisionErrors);
        else
            Adapter->TxAbortExcessCollisions += (pTxStats->TxExcessiveCollisionErrors + (0xFFFFFUL - Adapter->TxStats.TxExcessiveCollisionErrors) + 1UL);

        Adapter->TxStats.TxExcessiveCollisionErrors = pTxStats->TxExcessiveCollisionErrors;

        // Tx Late Collisions (20Bit Counter)
        if (pTxStats->TxLateCollisionErrors >= Adapter->TxStats.TxLateCollisionErrors)
            Adapter->TxLateCollisions += (pTxStats->TxLateCollisionErrors - Adapter->TxStats.TxLateCollisionErrors);
        else
            Adapter->TxLateCollisions += (pTxStats->TxLateCollisionErrors + (0xFFFFFUL - Adapter->TxStats.TxLateCollisionErrors) + 1UL);

        Adapter->TxStats.TxLateCollisionErrors = pTxStats->TxLateCollisionErrors;

        // Tx Underrun (20Bit Counter)
        if (pTxStats->TxBufferUnderrunErrors >= Adapter->TxStats.TxBufferUnderrunErrors)
            Adapter->TxDmaUnderrun += (pTxStats->TxBufferUnderrunErrors - Adapter->TxStats.TxBufferUnderrunErrors);
        else
            Adapter->TxDmaUnderrun += (pTxStats->TxBufferUnderrunErrors + (0xFFFFFUL - Adapter->TxStats.TxBufferUnderrunErrors) + 1UL);

        Adapter->TxStats.TxBufferUnderrunErrors = pTxStats->TxBufferUnderrunErrors;

        // Tx Carrier Errors (20Bit Counter)
        if (pTxStats->TxCarrierErrors >= Adapter->TxStats.TxCarrierErrors)
            Adapter->TxLostCRS += (pTxStats->TxCarrierErrors - Adapter->TxStats.TxCarrierErrors);
        else
            Adapter->TxLostCRS += (pTxStats->TxCarrierErrors + (0xFFFFFUL - Adapter->TxStats.TxCarrierErrors) + 1UL);

        Adapter->TxStats.TxCarrierErrors = pTxStats->TxCarrierErrors;

        // Tx Single Collision (20Bit Counter)
        if (pTxStats->TxSingleCollisions >= Adapter->TxStats.TxSingleCollisions)
            Adapter->OneRetry += (pTxStats->TxSingleCollisions - Adapter->TxStats.TxSingleCollisions);
        else
            Adapter->OneRetry += (pTxStats->TxSingleCollisions + (0xFFFFFUL - Adapter->TxStats.TxSingleCollisions) + 1UL);

        Adapter->TxStats.TxSingleCollisions = pTxStats->TxSingleCollisions;

        // Tx Multiple Collision (20Bit Counter)
        if (pTxStats->TxMultipleCollisions >= Adapter->TxStats.TxMultipleCollisions)
            Adapter->MoreThanOneRetry += (pTxStats->TxMultipleCollisions - Adapter->TxStats.TxMultipleCollisions);
        else
            Adapter->MoreThanOneRetry += (pTxStats->TxMultipleCollisions + (0xFFFFFUL - Adapter->TxStats.TxMultipleCollisions) + 1UL);

        Adapter->TxStats.TxMultipleCollisions = pTxStats->TxMultipleCollisions;

        // Tx Excessive Deferral (20Bit Counter)
        if (pTxStats->TxExcessiveDeferralErrors >= Adapter->TxStats.TxExcessiveDeferralErrors)
            Adapter->TransmitFailuresOther += (pTxStats->TxExcessiveDeferralErrors - Adapter->TxStats.TxExcessiveDeferralErrors);
        else
            Adapter->TransmitFailuresOther += (pTxStats->TxExcessiveDeferralErrors + (0xFFFFFUL - Adapter->TxStats.TxExcessiveDeferralErrors) + 1UL);

        Adapter->TxStats.TxExcessiveDeferralErrors = pTxStats->TxExcessiveDeferralErrors;

        // Tx Pause (20Bit Counter)
        if (pTxStats->TxPauseFrames >= Adapter->TxStats.TxPauseFrames)
            Adapter->TxPauseFrames += (pTxStats->TxPauseFrames - Adapter->TxStats.TxPauseFrames);
        else
            Adapter->TxPauseFrames += (pTxStats->TxPauseFrames + (0xFFFFFUL - Adapter->TxStats.TxPauseFrames) + 1UL);

        Adapter->TxStats.TxPauseFrames = pTxStats->TxPauseFrames;

        // Tx Bad (20Bit Counter)
        if (pTxStats->TxBadFrames >= Adapter->TxStats.TxBadFrames)
            Adapter->TxBadFrames += (pTxStats->TxBadFrames - Adapter->TxStats.TxBadFrames);
        else
            Adapter->TxBadFrames += (pTxStats->TxBadFrames + (0xFFFFFUL - Adapter->TxStats.TxBadFrames) + 1UL);

        Adapter->TxStats.TxBadFrames = pTxStats->TxBadFrames;
      
        // EEE Tx LPI Transitions(20Bit Counter)
        if (pTxStats->EeeTxLpiTransitions >= Adapter->TxStats.EeeTxLpiTransitions)
            Adapter->EeeTxLpiTransitions += (pTxStats->EeeTxLpiTransitions - Adapter->TxStats.EeeTxLpiTransitions);
        else
            Adapter->EeeTxLpiTransitions += (pTxStats->EeeTxLpiTransitions + (0xFFFFFUL - Adapter->TxStats.EeeTxLpiTransitions) + 1UL);
        
        Adapter->TxStats.EeeTxLpiTransitions = pTxStats->EeeTxLpiTransitions;
       
        // EEE Tx LPI Time (20Bit Counter)
        if (pTxStats->EeeTxLpiTime >= Adapter->TxStats.EeeTxLpiTime)
            Adapter->EeeTxLpiTime += (pTxStats->EeeTxLpiTime - Adapter->TxStats.EeeTxLpiTime);
        else
            Adapter->EeeTxLpiTime += (pTxStats->EeeTxLpiTime + (0xFFFFFUL - Adapter->TxStats.EeeTxLpiTime) + 1UL);
        
        Adapter->TxStats.EeeTxLpiTime = pTxStats->EeeTxLpiTime;

        // Good Rx (32Bit Counter)
        if (pRxStats->RxGoodFrames >= Adapter->RxStats.RxGoodFrames)
            Adapter->ReceivesGood += (ULONG64)(pRxStats->RxGoodFrames - Adapter->RxStats.RxGoodFrames);
        else
            Adapter->ReceivesGood += (ULONG64)(pRxStats->RxGoodFrames + (0xFFFFFFFFUL - Adapter->RxStats.RxGoodFrames) + 1UL);

        Adapter->RxStats.RxGoodFrames = pRxStats->RxGoodFrames;

        // Rx Crc (20Bit Counter)
        if (pRxStats->RxCrcErrors >= Adapter->RxStats.RxCrcErrors)  
            Adapter->RcvCrcErrors += (pRxStats->RxCrcErrors - Adapter->RxStats.RxCrcErrors);
        else
            Adapter->RcvCrcErrors += (pRxStats->RxCrcErrors + (0xFFFFFUL - Adapter->RxStats.RxCrcErrors) + 1UL);

        Adapter->RxStats.RxCrcErrors = pRxStats->RxCrcErrors;

        // Rx Alignment (20Bit Counter)
        if (pRxStats->RxAlignmentErrors >= Adapter->RxStats.RxAlignmentErrors)
            Adapter->RcvAlignmentErrors += (pRxStats->RxAlignmentErrors - Adapter->RxStats.RxAlignmentErrors);
        else
            Adapter->RcvAlignmentErrors += (pRxStats->RxAlignmentErrors + (0xFFFFFUL - Adapter->RxStats.RxAlignmentErrors) + 1UL);

        Adapter->RxStats.RxAlignmentErrors = pRxStats->RxAlignmentErrors;

        // Rx Resources (20Bit Counter)
        if (pRxStats->RxFifoDroppedFrames >= Adapter->RxStats.RxFifoDroppedFrames)
            Adapter->RcvResourceErrors += (pRxStats->RxFifoDroppedFrames - Adapter->RxStats.RxFifoDroppedFrames);
        else
            Adapter->RcvResourceErrors += (pRxStats->RxFifoDroppedFrames + (0xFFFFFUL - Adapter->RxStats.RxFifoDroppedFrames) + 1UL);

        Adapter->RxStats.RxFifoDroppedFrames = pRxStats->RxFifoDroppedFrames;

        // Rx Runts (20Bit Counter)
        if (pRxStats->RxRuntFrameErrors >= Adapter->RxStats.RxRuntFrameErrors)
            Adapter->RcvRuntErrors += (pRxStats->RxRuntFrameErrors - Adapter->RxStats.RxRuntFrameErrors);
        else
            Adapter->RcvRuntErrors += (pRxStats->RxRuntFrameErrors + (0xFFFFFUL - Adapter->RxStats.RxRuntFrameErrors) + 1UL);

        Adapter->RxStats.RxRuntFrameErrors = pRxStats->RxRuntFrameErrors;

        // Rx TooLong (20Bit Counter)
        if (pRxStats->RxFrameTooLongError >= Adapter->RxStats.RxFrameTooLongError)
            Adapter->RcvFrameTooLongErrors += (pRxStats->RxFrameTooLongError - Adapter->RxStats.RxFrameTooLongError);
        else
            Adapter->RcvFrameTooLongErrors += (pRxStats->RxFrameTooLongError + (0xFFFFFUL - Adapter->RxStats.RxFrameTooLongError) + 1UL);

        Adapter->RxStats.RxFrameTooLongError = pRxStats->RxFrameTooLongError;

        // Rx Late Collisions (20Bit Counter)
        if (pRxStats->RxLaterCollisionError >= Adapter->RxStats.RxLaterCollisionError)
            Adapter->RcvLateCollisionErrors += (pRxStats->RxLaterCollisionError - Adapter->RxStats.RxLaterCollisionError);
        else
            Adapter->RcvLateCollisionErrors += (pRxStats->RxLaterCollisionError + (0xFFFFFUL - Adapter->RxStats.RxLaterCollisionError) + 1UL);

        Adapter->RxStats.RxLaterCollisionError = pRxStats->RxLaterCollisionError;

        // Rx Bad (20Bit Counter)
        if (pRxStats->RxBadFrames >= Adapter->RxStats.RxBadFrames)
            Adapter->RcvBadErrors += (pRxStats->RxBadFrames - Adapter->RxStats.RxBadFrames);
        else
            Adapter->RcvBadErrors += (pRxStats->RxBadFrames + (0xFFFFFUL - Adapter->RxStats.RxBadFrames) + 1UL);

        Adapter->RxStats.RxBadFrames = pRxStats->RxBadFrames;

         // EEE Rx LPI Transitions(20Bit Counter)
        if (pRxStats->EeeRxLpiTransitions >= Adapter->RxStats.EeeRxLpiTransitions)
            Adapter->EeeRxLpiTransitions += (pRxStats->EeeRxLpiTransitions - Adapter->RxStats.EeeRxLpiTransitions);
        else
            Adapter->EeeRxLpiTransitions += (pRxStats->EeeRxLpiTransitions + (0xFFFFFUL - Adapter->RxStats.EeeRxLpiTransitions) + 1UL);
        
        Adapter->RxStats.EeeRxLpiTransitions = pRxStats->EeeRxLpiTransitions;

        // EEE Rx LPI Time (20Bit Counter)
        if (pRxStats->EeeRxLpiTime >= Adapter->RxStats.EeeRxLpiTime)
            Adapter->EeeRxLpiTime += (pRxStats->EeeRxLpiTime - Adapter->RxStats.EeeRxLpiTime);
        else
            Adapter->EeeRxLpiTime += (pRxStats->EeeRxLpiTime + (0xFFFFFUL - Adapter->RxStats.EeeRxLpiTime) + 1UL);
        
        Adapter->RxStats.EeeRxLpiTime = pRxStats->EeeRxLpiTime;
    }
    else
    {
        // Good Tx (32Bit Counter)
        Adapter->TransmitsGood += (ULONG64)pTxStats->TxGoodFrames;

        // Tx Excess Collisions (20Bit Counter)
        Adapter->TxAbortExcessCollisions += pTxStats->TxExcessiveCollisionErrors;

        // Tx Late Collisions (20Bit Counter)
        Adapter->TxLateCollisions += pTxStats->TxLateCollisionErrors;

        // Tx Underrun (20Bit Counter)
        Adapter->TxDmaUnderrun += pTxStats->TxBufferUnderrunErrors;

        // Tx Carrier Errors (20Bit Counter)
        Adapter->TxLostCRS += pTxStats->TxCarrierErrors;

        // Tx Single Collision (20Bit Counter)
        Adapter->OneRetry += pTxStats->TxSingleCollisions;

        // Tx Multiple Collision (20Bit Counter)
        Adapter->MoreThanOneRetry += pTxStats->TxMultipleCollisions;

        // Tx Excessive Deferral (20Bit Counter)
        Adapter->TransmitFailuresOther += pTxStats->TxExcessiveDeferralErrors;

        // Tx Pause (20Bit Counter)
        Adapter->TxPauseFrames += pTxStats->TxPauseFrames;

        // Tx Bad (20Bit Counter)
        Adapter->TxBadFrames += pTxStats->TxBadFrames;

        // EEE Tx LPI Transitions(20Bit Counter)
        Adapter->EeeTxLpiTransitions = pTxStats->EeeTxLpiTransitions;
        
        // EEE Tx LPI Time (20Bit Counter)       
        Adapter->EeeTxLpiTime = pTxStats->EeeTxLpiTime;

        // Good Rx (32Bit Counter)
        Adapter->ReceivesGood += (ULONG64)pRxStats->RxGoodFrames;

        // Rx Crc (20Bit Counter)
        Adapter->RcvCrcErrors += pRxStats->RxCrcErrors;

        // Rx Alignment (20Bit Counter)
        Adapter->RcvAlignmentErrors += pRxStats->RxAlignmentErrors;

        // Rx Resources (20Bit Counter)
        Adapter->RcvResourceErrors += pRxStats->RxFifoDroppedFrames;

        // Rx Runts (20Bit Counter)
        Adapter->RcvRuntErrors += pRxStats->RxRuntFrameErrors;

        // Rx TooLong (20Bit Counter)
        Adapter->RcvFrameTooLongErrors += pRxStats->RxFrameTooLongError;

        // Rx Late Collisions (20Bit Counter)
        Adapter->RcvLateCollisionErrors += pRxStats->RxLaterCollisionError;

        // Rx Bad (20Bit Counter)
        Adapter->RcvBadErrors += pRxStats->RxBadFrames;

        // EEE Rx LPI Transitions(20Bit Counter)
        Adapter->EeeRxLpiTransitions = pRxStats->EeeRxLpiTransitions;
        
        // EEE Rx LPI Time (20Bit Counter)       
        Adapter->EeeRxLpiTime = pRxStats->EeeRxLpiTime;
    }
}

/*++

Routine Description:

   Workitem to check link state. This workitem is scheduled in the link check 
   timer callback function 
	  
Arguments:

	WorkItem - Pointer to workitem
	
	Context - Pointer to our Adapter

Return Value:

	None.

--*/
VOID
NICLinkCheckWorkItemCallBack(
	PNDIS_WORK_ITEM WorkItem, 
	PVOID Context
	)
{
	NIC_STATUS		Status;
	DWORD			Value32;
	PMP_ADAPTER 	Adapter = (PMP_ADAPTER)Context;
	UCHAR			Stats[sizeof(LAN9500_RX_STATS) + sizeof(LAN9500_TX_STATS)];
	
	UNREFERENCED_PARAMETER(WorkItem);

	SMSC_TRACE0(DBG_LINK,"--> NICLinkCheckWorkItemCallBack\r\n");

	if (!MP_TEST_FLAG(Adapter, fMP_PERIODIC_LINK_CHECK) || !MP_IS_READY(Adapter))
		goto Exit_NICLinkCheckWorkItemCallBack;

	if(!NICIsCardPresent(Adapter))
		goto Exit_NICLinkCheckWorkItemCallBack;

	Status = NICReadStats(Adapter, Stats);
	
	if (Status == NIC_STATUS_SUCCESS)
	{
		NICUpdateStatCounters(Adapter, Stats);
	}
	else
	{
		SMSC_TRACE1(DBG_LINK,"NICLinkCheckWorkItemCallBack. Error reading stats (0x%08lx)\r\n", Status);
	}

    if (Adapter->MdioDisable & MDIO_DISABLE_INTERNAL)
    {
        // Do not actually check the link, only using this
        // timer to periodically read stats.
        goto Exit_NICLinkCheckWorkItemCallBackAfterElapsedCountersUpdates;
    }		
	// Check Link
	if(Adapter->PerformingNWAY)
	{
		SMSC_TRACE1(DBG_LINK,"NWAY elapsed time = %d seconds since first check\r\n", Adapter->PerformingNWAYTimerCounter);
		if(NICIsNWAYComplete(Adapter) || (Adapter->PerformingNWAYTimerCounter == 3))
		{
			Status = NICServicePhyInterrupt(Adapter); // returns Adapter->PerformingNWAY = FALSE only if linked
			Adapter->PerformingNWAY = FALSE;

			NdisAcquireSpinLock(&Adapter->RequestLock);
/*
			if (Adapter->HardwareStatus == NdisHardwareStatusReset)
			{
				SMSC_TRACE1(DBG_RESET,"Reset status1= 0x%x \r\n", (!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) ? NDIS_STATUS_SUCCESS : NDIS_STATUS_ADAPTER_NOT_READY);
				NdisMResetComplete(
									Adapter->AdapterHandle, 
									(!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) ? NDIS_STATUS_SUCCESS : NDIS_STATUS_ADAPTER_NOT_READY,
									FALSE
								  );
			}
*/
			Adapter->HardwareStatus = NdisHardwareStatusReady;
			NdisReleaseSpinLock(&Adapter->RequestLock); 
		}

		Adapter->PerformingNWAYTimerCounter++;
	}


     Status = NICReadPhy(Adapter, PHY_THERA_MODE_CTL_STS, &Value32);
     if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICLinkCheckWorkItemCallBack;
     SMSC_TRACE1(DBG_LINK,"PHY_THERA_MODE_CTL_STS=0x%08lx\r\n", Value32);
     // Let's see if we need to put the PHY in EDPD
     if (MP_TEST_FLAG(Adapter, fMP_DISCONNECTED) && (!(Value32 & PHY_THERA_MODE_CTL_STS_ED_)) 
        && (Adapter->SecondsElapsedAfterMPInitializeExit >4) && (Adapter->EDPDNetDetachTimer++ == Adapter->EDPDNetDetachTimeout))
         {

                        if (Adapter->SleepPhyOnLinkDown && (Adapter->PhyHookup == PHY_HOOKUP_INTERNAL))
                        {
                                SMSC_TRACE1(DBG_INIT,"PHY_THERA_MODE_CTL_STS=0x%08lx\r\n", Value32);
                                SMSC_TRACE0(DBG_SHUT_DOWN,"Setting PHY into EDPD\r\n");

                                /*
                                                * Put the PHY in ED PD so that we can save some power.
                                                */

                                // If we are doing NWAY turn it off now
/*
                                if(Adapter->MediaType <= MEDIA_TYPE_AUTO_CUSTOM)
                                {
                                        Status = NICWritePhy(Adapter, PHY_BMC_REG, 0);
                                        if (Status != NIC_STATUS_SUCCESS)
                                                   goto Exit_NICLinkCheckWorkItemCallBack;
                                        SMSC_TRACE0(DBG_SHUT_DOWN,"Turn off NWAY\r\n");
                                }

                                // Turn off Auto MDIX
                                Status = NICReadPhy(Adapter, PHY_THERA_CTL_STS_IND, &Value32);
                                if (Status != NIC_STATUS_SUCCESS)
                                           goto Exit_NICLinkCheckWorkItemCallBack;
                                                
                                Value32 &= 0x1FFF;              
                                Status = NICWritePhy(Adapter, PHY_THERA_CTL_STS_IND, Value32);
                                if (Status != NIC_STATUS_SUCCESS)
                                           goto Exit_NICLinkCheckWorkItemCallBack;

                                SMSC_TRACE0(DBG_SHUT_DOWN,"Turn off AMDIX\r\n");
                                Adapter->SaveAutoMdix = Value32;


*/
                                // Setting up EDPD configuration in reg 16
                                if (Adapter->EDPDConfig != 0)
                                {
                                        Status = NICReadPhy(Adapter, PHY_THERA_SIL_REV, &Value32);
                                        if (Status != NIC_STATUS_SUCCESS)
                                                   goto Exit_NICLinkCheckWorkItemCallBack;
                                        SMSC_TRACE1(DBG_SHUT_DOWN,"Configure EDPD parameters. Writing 0x%04x to Reg 16\r\n",
                                                                                                (Value32 & 0x03FE) | Adapter->EDPDConfig);
                                        Status = NICWritePhy(Adapter, PHY_THERA_SIL_REV, (Value32 & 0x03FE) | Adapter->EDPDConfig);
                                        if (Status != NIC_STATUS_SUCCESS)
                                                   goto Exit_NICLinkCheckWorkItemCallBack;
                                }

                                Status = NICSetPHYInterruptCapability(Adapter, PHY_THERA_INT_ENERGY_ON);
                                if (Status != NIC_STATUS_SUCCESS)
                                           goto Exit_NICLinkCheckWorkItemCallBack;
                                
                                // Clear the WUPS bits to avoid spurious wake up
                                Status = NICReadRegister(Adapter, ETH_PM_CTRL, &Value32);
                                if (Status != NIC_STATUS_SUCCESS)
                                           goto Exit_NICLinkCheckWorkItemCallBack;
                                                
                                Value32 |= PM_CTL_WUPS_MULTI_;                          
                                Status = NICWriteRegister(Adapter, ETH_PM_CTRL, &Value32 );
                                if (Status != NIC_STATUS_SUCCESS)
                                           goto Exit_NICLinkCheckWorkItemCallBack;

                                // Set phy in ED low power mode.
                                Status = NICReadPhy(Adapter, PHY_THERA_MODE_CTL_STS, &Value32);
                                if (Status != NIC_STATUS_SUCCESS)
                                           goto Exit_NICLinkCheckWorkItemCallBack;
                                SMSC_TRACE0(DBG_SHUT_DOWN,"Putting PHY into EDPD Now ...\r\n");
                                SMSC_TRACE1(DBG_SHUT_DOWN,"PHY_THERA_MODE_CTL_STS=0x%08lx\r\n", Value32 | PHY_THERA_MODE_CTL_STS_EPD_);
                                                
                                Value32 |= PHY_THERA_MODE_CTL_STS_EPD_;
                                Status = NICWritePhy(Adapter, PHY_THERA_MODE_CTL_STS, Value32 );
                                if (Status != NIC_STATUS_SUCCESS)
                                           goto Exit_NICLinkCheckWorkItemCallBack;

                                Adapter->PhyLinkPowerDown = TRUE;
                        }

                        // Let's see if we need to do Smartdetach

                        if (Adapter->DetachOnLinkDownMode == DOLD_ENABLED)      {
                                        
                            SMSC_TRACE0(DBG_SHUT_DOWN,"Setting up Smart Detach\r\n");

                            if (!Adapter->SleepPhyOnLinkDown && (Adapter->PhyHookup == PHY_HOOKUP_INTERNAL))
                            {
                                 // Phy interrupts as reconnect source ?
                                 if (Adapter->LinkWakeModes & 0xFF)
                                 {
                                        // Clear any pending Phy interrupt and enable the mask for it
                                        SMSC_TRACE0(DBG_SHUT_DOWN,"Enabling PHY interrupts as reconnect Source\r\n");

                                        Status = NICSetPHYInterruptCapability(Adapter, ((WORD)(Adapter->LinkWakeModes & 0xFFFF)));

                                        // Clear the WUPS bits to avoid spurious wake up
                                        Status = NICReadRegister(Adapter, ETH_PM_CTRL, &Value32);
                                        if (Status != NIC_STATUS_SUCCESS)
                                                   goto Exit_NICLinkCheckWorkItemCallBack;
                                                                
                                        Value32 |= PM_CTL_WUPS_MULTI_;          
                                        Status = NICWriteRegister(Adapter, ETH_PM_CTRL, &Value32);
                                        if (Status != NIC_STATUS_SUCCESS)
                                                   goto Exit_NICLinkCheckWorkItemCallBack;

                                        // Set all supported media in NWAY Adver. Register in case it's disabled by forced link
                                        Status = NICReadPhy(Adapter, PHY_ANA_REG, &Value32);
                                           if (Status != NIC_STATUS_SUCCESS)
                                                      goto Exit_NICLinkCheckWorkItemCallBack;
                                        Value32 |= PHY_ANA_ALL_SUPPORTED_MEDIA;                 
                                           Status = NICWritePhy(Adapter, PHY_ANA_REG, Value32);
                                           if (Status != NIC_STATUS_SUCCESS)
                                                      goto Exit_NICLinkCheckWorkItemCallBack;
                                        
                                        // Enable NWAY in PHY_BMC in case it's disabled by forced link
                                        Status = NICWritePhy(Adapter, PHY_BMC_REG, PHY_AN_ENABLE | PHY_RESTART_AN);
                                 }

                                 // Link interrupt as reconnect source ?
                                 if (Adapter->LinkWakeModes & ETH_GPIO_WAKE_PHY_LINKUP_EN_)
                                 {
                                        SMSC_TRACE1(DBG_SHUT_DOWN,"Enabling Link on GPIO%d as reconnect Source\r\n",
                                                                Adapter->SDReconnectGpio == 11 ? 7 : Adapter->SDReconnectGpio);

                                        // Setup Linkup for either internal (GPIO7) or external GPIO that carries link.
                                        // Do it while preserving values for every other bit in this register.
                                        Status = NICReadRegister(Adapter, ETH_GPIO_WAKE, &Value32);
                                        if (Adapter->SDReconnectGpio == 11)
                                                // Enable wake on link for GPIO7 with polarity set to "driven low"
                                                Value32 = ( 
                                                                        ETH_GPIO_WAKE_PHY_LINKUP_EN_ |
                                                                        (Value32 & (~ETH_GPIO_WAKE_GPIOPOL7_)) | 
                                                                        ETH_GPIO_WAKE_GPIOWK7_ 
                                                                );
                                        else
                                        /*
                                                                      * Enable wake on GPIO<n> with polarity set to "driven low"
                                                                      */

                                        // Clear wake on link and polarity bit for this GPIO.
                                        Value32 &= (~(ETH_GPIO_WAKE_PHY_LINKUP_EN_ | (ETH_GPIO_WAKE_GPIOPOL0_ << Adapter->SDReconnectGpio)));

                                        // Enable GPIO for wakes and set the registry defined polarity for it.
                                        Value32 |= ((ETH_GPIO_WAKE_GPIOWK0_ | (Adapter->SDReconnectGpioPol << 16) ) << Adapter->SDReconnectGpio); 

                                        Status = NICWriteRegister(Adapter, ETH_GPIO_WAKE, &Value32);
                                        if (Status != NIC_STATUS_SUCCESS)
                                                   goto Exit_NICLinkCheckWorkItemCallBack;

                                        // Set GPIO7 as input
                                        Status = NICReadRegister( Adapter, ETH_GPIO_CFG, &Value32);
                                        if (Status != NIC_STATUS_SUCCESS)
                                                   goto Exit_NICLinkCheckWorkItemCallBack;
                                                
                                        if (Adapter->SDReconnectGpio == 11)
                                        {
                                                // Enable GPIO7 as input; data and type are don't care.
                                                Value32 &= (~((ETH_GPIO_CFG_GPO0_DISABLED_ | ETH_GPIO_CFG_GPO0_DIR_) << 7));  
                                        }
                                        else
                                        {
                                                // Enable GPIO<n> as input; data and type are don't care.
                                                Value32 &= (~((ETH_GPIO_CFG_GPO0_DISABLED_ | ETH_GPIO_CFG_GPO0_DIR_) << Adapter->SDReconnectGpio)); 
                                        }

                                        Status = NICWriteRegister( Adapter, ETH_GPIO_CFG, &Value32);
                                   }
                            }
                                
                            if (Status != NIC_STATUS_SUCCESS)
                                   goto Exit_NICLinkCheckWorkItemCallBack;

                        
                            Status = NICReadPhy(Adapter, PHY_THERA_MODE_CTL_STS, &Value32);
                            if (Status != NIC_STATUS_SUCCESS)
                                   goto Exit_NICLinkCheckWorkItemCallBack;

                            SMSC_TRACE1(DBG_INIT,"PHY_THERA_MODE_CTL_STS1=0x%08lx\r\n", Value32);
                
                            if (!(Value32 & PHY_THERA_MODE_CTL_STS_ED_))
                            {

				                //Before NetDetach respond to any pending requests
				                NICCompleteOidQueryIfPending(Adapter);

                                SMSC_TRACE0(DBG_SHUT_DOWN,"Enabling Smart Detach now...\n");

                                // Now do the Smart Detach !
                                // USB bus will be disconnected as soon as ETH_HW_CFG_SMDET_EN_ bit is set. 
                                // So from this point all USB transactions will fail including the following register write.
                             
                                Status = NICReadRegister(Adapter, ETH_HW_CFG, &Value32);
                                if (Status != NIC_STATUS_SUCCESS)
                                        goto Exit_NICLinkCheckWorkItemCallBack;
                             
                                Value32 |= ETH_HW_CFG_SMDET_EN_;
                                Status = NICWriteRegister(Adapter, ETH_HW_CFG, &Value32 );
                                if (Status != NIC_STATUS_SUCCESS)
                                {
                                    // At this point we don't know if NICWriteRegister failed for any reason other than Smart Detach.
                                    // Let's try the following read to make sure USB bus is really detached
                                    Status = NICReadRegister(Adapter, ETH_HW_CFG, &Value32);
                                    if (Status != NIC_STATUS_SUCCESS)
                                    {
                                        Status = NIC_STATUS_SUCCESS; // sucessfully detached
                                    }
                                    else
                                    {
                                        SMSC_TRACE0(DBG_ERROR,"USB bus is still attached after setting Smart Detach Bit!!!\n");
                                        SMSC_TRACE1(DBG_ERROR,"ETH_HW_CFG=0x%08lX\n", Value32);
                                        Status = NIC_STATUS_FAILURE; // Detach failed
                                    }
                                    goto Exit_NICLinkCheckWorkItemCallBack;
                                                                                                        
                                }
                           }
                     }
     }

	// If we are disconnected and EDPD is enabled, see if we need to 
	// come out of EDPD
	if ((MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && (Adapter->PhyLinkPowerDown == TRUE))
	{
		Status = NICReadPhy(Adapter, PHY_THERA_MODE_CTL_STS, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICLinkCheckWorkItemCallBack;
		SMSC_TRACE1(DBG_INIT,"PHY_THERA_MODE_CTL_STS2=0x%08lx\r\n", Value32);

		if (Value32 & PHY_THERA_MODE_CTL_STS_ED_)
		{
			SMSC_TRACE0(DBG_INIT,"Energy Detected !!! Turning OFF EDPD\r\n");

			Value32 &= ~PHY_THERA_MODE_CTL_STS_EPD_;
			Status = NICWritePhy(Adapter, PHY_THERA_MODE_CTL_STS, Value32);
			if (Status != NIC_STATUS_SUCCESS)
				goto Exit_NICLinkCheckWorkItemCallBack;

			Adapter->PhyLinkPowerDown = FALSE;
/*
			if (Adapter->PhyHookup == PHY_HOOKUP_INTERNAL)
			{
				// Restore Auto MDIX
				Status = NICWritePhy(Adapter, PHY_THERA_CTL_STS_IND, Adapter->SaveAutoMdix);
				if(Status != NIC_STATUS_SUCCESS)
					goto Exit_NICLinkCheckWorkItemCallBack;
				SMSC_TRACE0(DBG_LINK,"AMDIX Restored\r\n");
			}
*/
			// If we were doing NWAY turn it ON now
			if(Adapter->MediaType <= MEDIA_TYPE_AUTO_CUSTOM)
			{
				Status = NICWritePhy(  
										Adapter, 
										PHY_BMC_REG, 
										PHY_AN_ENABLE
									 );
				if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICLinkCheckWorkItemCallBack;
	
				SMSC_TRACE0(DBG_LINK,"NWAY Reenabled\r\n");
			}


            /* After 3 sec, go to EDPD. For Gig speed it takes longer to do auto-neg. After this
			   timeout and if no link the device need to be put to EDPD.
			*/	
			Adapter->EDPDNetDetachTimer = 0;
			Adapter->EDPDNetDetachTimeout = 3; 
		}
	}

	// The Phy is still powered down, skip all link checks 
	if (Adapter->PhyLinkPowerDown == TRUE)
		goto Exit_NICLinkCheckWorkItemCallBack;

	// Check to see if the media was reconnected to workaround problem of 
	//	no int generated on link reconnection when NWAY is off.

	if(
		!Adapter->PerformingNWAY &&
		(Adapter->MediaType > MEDIA_TYPE_AUTO_CUSTOM) 
	  )
	{
		DWORD DirectAttachedPhyLinkStatus = 0;
		DWORD OtherPhysLinkStatus = 0;

		Status = NICReadPhy(Adapter, PHY_BMS_REG, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICLinkCheckWorkItemCallBack;

		Status = NICReadPhy(Adapter, PHY_BMS_REG, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICLinkCheckWorkItemCallBack;

		if( (MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && (Value32 & PHY_LINK) )
		{
			DirectAttachedPhyLinkStatus = 1;
			//MP_CLEAR_FLAG(Adapter, fMP_DISCONNECTED);
			//NICIndicateMediaStatus(Adapter);
		}
		//When LAN9500 is connected to switch (say, LAN9303) the link status of
		//other external ports can be included in the link status reporting to NDIS.
		//Ex for LAN9303: Link status: Port0 && (port1 || port2)
		//For: LAN9303: MdioPhyAddressStart registry must be set to 1 (port1) and 
		//MdioPhyAddressCount registry should be set to 2 (last phy address, port2).
		if (DirectAttachedPhyLinkStatus)
		{
			if (Adapter->MdioPhyAddressCount)
			{
				Status = NICCheckOtherPhysLinkStatus(Adapter, &OtherPhysLinkStatus);
				if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICLinkCheckWorkItemCallBack;
				if (DirectAttachedPhyLinkStatus && OtherPhysLinkStatus) {
					MP_CLEAR_FLAG(Adapter, fMP_DISCONNECTED);
					SMSC_TRACE0(DBG_LINK,"NICLinkCheckWorkItemCallBack: OtherPhysLinkStatus=1\r\n");
					NICIndicateMediaStatus(Adapter);
				}
			} else {
				MP_CLEAR_FLAG(Adapter, fMP_DISCONNECTED);
				NICIndicateMediaStatus(Adapter);
			}
		}
	}

	// If the hardware status was not ready, and we are now then switch it to ready.

	NdisAcquireSpinLock(&Adapter->RequestLock);
	if(
		(!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && (!MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS)) &&
		(
			(Adapter->HardwareStatus == NdisHardwareStatusInitializing) ||
			(Adapter->HardwareStatus == NdisHardwareStatusReset)
		)
	  )
	{
	/*
		if (Adapter->HardwareStatus == NdisHardwareStatusReset)
		{
			SMSC_TRACE1(DBG_RESET,"Reset status2= 0x%x \r\n", (!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) ? NDIS_STATUS_SUCCESS : NDIS_STATUS_ADAPTER_NOT_READY);
			NdisMResetComplete(
								Adapter->AdapterHandle, 
								(!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) ? NDIS_STATUS_SUCCESS : NDIS_STATUS_ADAPTER_NOT_READY,
								FALSE
							  );

		}
	*/
		SMSC_TRACE0(DBG_LINK,"Changing Hardware Status (OID_GEN_HARDWARE_STATUS) to ready\r\n");
		Adapter->HardwareStatus = NdisHardwareStatusReady;
	}
	NdisReleaseSpinLock(&Adapter->RequestLock); 

	// If two seconds elapsed since returning from MPInitialize and we still
	// don't have link, indicate a media disconnect event.

	if (
		(Adapter->SecondsElapsedAfterMPInitializeExit == 2) &&
		(MP_TEST_FLAG(Adapter, fMP_DISCONNECTED))
	   )
	{
		NICIndicateMediaStatus(Adapter);
	}

	// If six seconds elapsed since returning from MP_Reset with PENDING status and we still
	// don't have link, indicate a reset completion with the adapter not ready.

	NdisAcquireSpinLock(&Adapter->RequestLock);
	if (
		(Adapter->HardwareStatus == NdisHardwareStatusReset) &&
		(Adapter->SecondsElapsedAfterMPResetExit == 6) &&
		(MP_TEST_FLAG(Adapter, fMP_DISCONNECTED))
	   )
	{
		Adapter->HardwareStatus = NdisHardwareStatusNotReady;
		/*
		SMSC_TRACE1(DBG_RESET,"Reset status3= 0x%x \r\n",  NDIS_STATUS_ADAPTER_NOT_READY);
		NdisMResetComplete(
							Adapter->AdapterHandle, 
							NDIS_STATUS_ADAPTER_NOT_READY,
							FALSE
						  );
		*/

	}
	NdisReleaseSpinLock(&Adapter->RequestLock); 

	// In case the phy does not support it or we miss the link 
	// disconnection interrupt do a simple check for link change
	Status = NICReadPhy(Adapter, PHY_BMS_REG, &Value32);
	if (Status != NIC_STATUS_SUCCESS)
		goto Exit_NICLinkCheckWorkItemCallBack;

	Status = NICReadPhy(Adapter, PHY_BMS_REG, &Value32);
	if (Status != NIC_STATUS_SUCCESS)
		goto Exit_NICLinkCheckWorkItemCallBack;

	if (!Adapter->MdioPhyAddressCount) {
		if ((!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && !(Value32 & PHY_LINK) )
		{
			NICServicePhyInterrupt(Adapter);
		}
		else if ((MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && (Value32 & PHY_LINK) )
		{
			NICServicePhyInterrupt(Adapter);
		}
	} else {
		DWORD DirectAttachedPhyLinkStatus = 0;
		DWORD OtherPhysLinkStatus = 0;

		if (Value32 & PHY_LINK) {
			DirectAttachedPhyLinkStatus = 1;
		}
		Status = NICCheckOtherPhysLinkStatus(Adapter, &OtherPhysLinkStatus);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICLinkCheckWorkItemCallBackAfterElapsedCountersUpdates;
		if ((MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && DirectAttachedPhyLinkStatus && OtherPhysLinkStatus)
		{
			SMSC_TRACE0(DBG_LINK,"NICServicePhyInterrupt: OtherPhysLinkStatus=1, Indicating mediaconnection\r\n");
			MP_CLEAR_FLAG(Adapter, fMP_DISCONNECTED);
			NICIndicateMediaStatus(Adapter);
		}

		if ((!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && (!(DirectAttachedPhyLinkStatus && OtherPhysLinkStatus)))
		{
			SMSC_TRACE0(DBG_LINK,"NICServicePhyInterrupt: OtherPhysLinkStatus=1, Indicating Media Dis-connection\r\n");
			MP_SET_FLAG(Adapter, fMP_DISCONNECTED);
			NICIndicateMediaStatus(Adapter);
		}
	}
	// Is there still a pending Oid Request we have not completed and maybe could ?
	if (!Adapter->PerformingNWAY && !MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS))
	{
		// Complete it !
		if (MP_TEST_FLAG(Adapter, fMP_REQUEST_PENDING))
			NICCompleteOidQueryIfPending(Adapter);
	}

Exit_NICLinkCheckWorkItemCallBack:
	// Update counters

	Adapter->SecondsElapsedAfterMPInitializeExit++;
	// Take care of the case when the counter wraps around to 0.
	if (Adapter->SecondsElapsedAfterMPInitializeExit == 0)
	{
		Adapter->SecondsElapsedAfterMPInitializeExit = 3;
	}

	Adapter->SecondsElapsedAfterMPResetExit++;
	// Take care of the case when the counter wraps around to 0.
	if (Adapter->SecondsElapsedAfterMPResetExit == 0)
	{
		Adapter->SecondsElapsedAfterMPResetExit = 3;
	}

Exit_NICLinkCheckWorkItemCallBackAfterElapsedCountersUpdates:

	MP_CLEAR_FLAG(Adapter, fMP_LINK_CHECK_IN_PROGRESS);

	//
	// Clear the flag to let the WorkItem structure be reused for
	// scheduling another one.
	//
	InterlockedExchange(&Adapter->IsLinkCheckWorkItemQueued, FALSE);

	MP_DEC_REF(Adapter);	 

	SMSC_TRACE0(DBG_LINK,"<-- NICLinkCheckWorkItemCallBack\r\n");
}


/*++

Routine Description:

   This routine services the Phy Interrupt
	  
Arguments:

	Adapter    - pointer to the MP_ADAPTER structure
	
Return Value:

	NIC_STATUS_XXX

--*/
NIC_STATUS 
NICServicePhyInterrupt(
	PMP_ADAPTER Adapter
	)
{
	NIC_STATUS Status;
	DWORD Value32,Value32_2,Temp32;

	SMSC_TRACE0(DBG_PHY,"--> NICServicePhyInterrupt\r\n");

	// If by any chance we get here when using a serial PHY just go out
	// without attempting to talk to an MII PHY.

	// Reading BMS serves the purpose of seeing if we need to update 
	// link and speed in the device, and do media connection indication.
	// Actually read it twice, as sometimes the Link does not get updated
	// until the second read in some PHYs that latch the link down condition.

	Status = NICReadPhy(Adapter, PHY_BMS_REG, &Temp32);
	if (Status != NIC_STATUS_SUCCESS)
		goto Exit_NICServicePhyInterrupt;

    Nic_phy_boost(Adapter,1);

	Status = NICReadPhy(Adapter, PHY_BMS_REG, &Temp32);
	if (Status != NIC_STATUS_SUCCESS)
		goto Exit_NICServicePhyInterrupt;

	if( (Adapter->MediaType <= MEDIA_TYPE_AUTO_CUSTOM) && 
		(Temp32 & PHY_LINK) )
	{
		// If NWAY was in process then signal we are done
		Adapter->PerformingNWAY = FALSE;

		// Get link speed 
		Status = NICReadPhy(Adapter, PHY_ANA_REG, &Value32_2);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICServicePhyInterrupt;

		Status = NICReadPhy(Adapter, PHY_ANLPA_REG, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICServicePhyInterrupt;

		// Get the least common denominator
		Value32 &= Value32_2;

		// Set the flags before calling NICSetMediaType.
		Adapter->LinkSpeed = LINK_SPEED_10MBS;
		if(Value32 & (PHY_100_BASE_TX_FD | PHY_100_BASE_TX_HD)) 
		{
			Adapter->LinkSpeed = LINK_SPEED_100MBS;
			SMSC_TRACE0(DBG_PHY,"Link Speed = 100Mbs\r\n");
		} 
		else 
		{
			if(Value32 & (PHY_10_BASE_T_FD | PHY_10_BASE_T_HD))
			{
				SMSC_TRACE0(DBG_PHY,"Link Speed = 10Mbs\r\n");
			}
			else
			{
				SMSC_TRACE0(DBG_PHY,"Link Speed = Unknown\r\n");
			}
		}

		// Get link duplex mode 
		Adapter->LinkDuplex = LINK_DUPLEX_HALF;
		if(Value32 & (PHY_100_BASE_TX_FD | PHY_10_BASE_T_FD)) 
		{
			Adapter->LinkDuplex = LINK_DUPLEX_FULL;
			SMSC_TRACE0(DBG_PHY,"Duplex = Full\r\n");
		} 
		else 
		{
			if(Value32 & (PHY_100_BASE_TX_HD | PHY_10_BASE_T_HD))
			{
				SMSC_TRACE0(DBG_PHY,"Duplex = Half\r\n");
			}
			else 
			{
				SMSC_TRACE0(DBG_PHY,"Duplex = Unknown\r\n");
			}
		}

		Status = NICSetMediaType(Adapter);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICServicePhyInterrupt;
	}

	// Signal Media connection status change if necessary, but only if NWAY is not
	// currently in progress.
	if (!Adapter->PerformingNWAY)
	{
		if (!Adapter->MdioPhyAddressCount) {
			if( (MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && (Temp32 & PHY_LINK) )
			{
				MP_CLEAR_FLAG(Adapter, fMP_DISCONNECTED);
				NICIndicateMediaStatus(Adapter);
			}

			if( (!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && (!(Temp32 & PHY_LINK)) )
			{
				MP_SET_FLAG(Adapter, fMP_DISCONNECTED);
				NICIndicateMediaStatus(Adapter);
			}
		} else {
			DWORD DirectAttachedPhyLinkStatus = 0;
			DWORD OtherPhysLinkStatus = 0;

			if (Temp32 & PHY_LINK) {
				DirectAttachedPhyLinkStatus = 1;
			}
			Status = NICCheckOtherPhysLinkStatus(Adapter, &OtherPhysLinkStatus);
			if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICServicePhyInterrupt;
			if ((MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && DirectAttachedPhyLinkStatus && OtherPhysLinkStatus)
			{
				SMSC_TRACE0(DBG_PHY,"NICServicePhyInterrupt: OtherPhysLinkStatus = 1, Indicating mediaconnection\r\n");
				MP_CLEAR_FLAG(Adapter, fMP_DISCONNECTED);
				NICIndicateMediaStatus(Adapter);
			}

			if ((!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && (!(DirectAttachedPhyLinkStatus && OtherPhysLinkStatus)))
			{
				SMSC_TRACE0(DBG_PHY,"NICServicePhyInterrupt: OtherPhysLinkStatus = 1, Indicating Media Dis-connection\r\n");
				MP_SET_FLAG(Adapter, fMP_DISCONNECTED);
				NICIndicateMediaStatus(Adapter);
			}
		}
	}

	Status = NIC_STATUS_SUCCESS;

Exit_NICServicePhyInterrupt:
	SMSC_TRACE0(DBG_PHY,"<-- NICServicePhyInterrupt\r\n");
	return Status;
}

NIC_STATUS
NICCheckOtherPhysLinkStatus(
	PMP_ADAPTER Adapter,
	int *OtherPhysLinkStatus
	)
{
	DWORD Count, Value32;
	NIC_STATUS Status = NIC_STATUS_SUCCESS;

	SMSC_TRACE0(DBG_PHY, "--> NICCheckOtherPhysLinkStatus\r\n");

	*OtherPhysLinkStatus = 0;
	if (Adapter->MdioPhyAddressCount)
	{
		//Read all the speicied phys for link status
		for (Count = Adapter->MdioPhyAddressStart; Count <= Adapter->MdioPhyAddressCount; Count++)
		{
			Status = NICReadPhyAddress(Adapter, Count, PHY_BMS_REG, &Value32, 0);
			if (Status != NIC_STATUS_SUCCESS)
				goto Exit_NICCheckOtherPhysLinkStatus;
			//Read again LL
			Status = NICReadPhyAddress(Adapter, Count, PHY_BMS_REG, &Value32, 0);
			if (Status != NIC_STATUS_SUCCESS)
				goto Exit_NICCheckOtherPhysLinkStatus;
			if(Value32 & PHY_LINK)
			{
				*OtherPhysLinkStatus = 1;
				//Any one of the external phy link up then break
				break;
			} else {
				*OtherPhysLinkStatus = 0;
			}
		}
	}
Exit_NICCheckOtherPhysLinkStatus:
	SMSC_TRACE1(DBG_PHY, "<-- NICCheckOtherPhysLinkStatus, OtherPhysLinkStatus = %d\r\n", *OtherPhysLinkStatus);
	return Status;
}
