/*++

	Copyright (c) 2008-2009 SMSC. All rights reserved.

	The information contained in this file is confidential and proprietary to
	Standard Microsystems Corporation.	No part of this file may be reproduced
	or distributed, in any form or by any means for any purpose, without the
	express written permission of Standard Microsystems Corporation.


Module Name:

	send.c
		
Abstract:

	This module contains miniport functions for handling Send
	packets and other helper routines called by the miniport function.
	
Revision History:
	12/11/2008	Initial version.

Notes:

--*/
#include "smsc9500.h"


/*++

Routine Description:

	This routine Starts the Tx path at the MAC and sets the flag to accept
	bulk out requests

Arguments:

	Adapter - pointer to our adapter context structure

Return Value:
	
	NIC_STATUS_XXX
	
--*/

NIC_STATUS
NICStartTxPath(
	PMP_ADAPTER Adapter
	)
{ 
	DWORD Value32,dwTemp;
	NIC_STATUS nStatus = NIC_STATUS_SUCCESS;
	 
	SMSC_TRACE0(DBG_TX,"--> NICStartTxPath\r\n");

	// Enable Tx at MAC
	nStatus=NICReadRegister(Adapter, MAC_CR, &Value32);
	Value32 |= MAC_CR_TXEN_;
	nStatus=NICWriteRegister(Adapter, MAC_CR, &Value32);
	if (nStatus !=NIC_STATUS_SUCCESS)
		goto Exit_NICStartTxPath;

	// Enable Tx at SCSRs
	dwTemp=ETH_TX_ON_;
	nStatus=NICWriteRegister(Adapter, ETH_TX_CFG, &dwTemp);
	
	if (nStatus !=NIC_STATUS_SUCCESS)
			goto Exit_NICStartTxPath;

	// We need to reset the pipe to make sure the data toggles are in sync.
	if(!NICResetPipe(Adapter, Adapter->BulkOut)) 
	{
		SMSC_TRACE0(DBG_ERROR,"Bulk Out Pipe Reset failed !!! Aborting Bulk Out Start sequence.\r\n");
		// Disable Tx at MAC
		nStatus=NICReadRegister(Adapter, MAC_CR, &Value32);
		Value32=Value32 & ~MAC_CR_TXEN_;
		nStatus=NICWriteRegister(Adapter, MAC_CR, &Value32);

		// Disable Tx at SCSRs
		dwTemp=ETH_TX_STOP_ | ETH_TX_FIFO_FLUSH_;
		nStatus=NICWriteRegister(Adapter, ETH_TX_CFG, &dwTemp);
	
		nStatus = NIC_STATUS_FAILURE;
		goto Exit_NICStartTxPath;
	}

	Adapter->BulkOut.PipeTransferAborted=0;
	NdisSetEvent(&Adapter->IsSendsCompletionWorkItemReady);

	// Set flag so that we can accept bulk out requests
	MP_SET_FLAG(Adapter, fMP_POST_WRITES);

	NdisSetEvent(&Adapter->SendsWaitingWorkItemReady);
	Adapter->IsSendReady=TRUE;

Exit_NICStartTxPath:
	SMSC_TRACE0(DBG_TX,"<-- NICStartTxPath\r\n");
	
	return (nStatus);
}


/*++

Routine Description:

   Workitem to post all the waiting queued sends. This workitem is scheduled 
   from the SendsCompletionWorkItemCallBack if there're packets to send in the 
   send wait queue.
	  
Arguments:

	WorkItem - Pointer to workitem
	
	Context - pointer to Adapter.

Return Value:

	None.

--*/

VOID
NICPostSendsWaitingWorkItemCallBack(
	PNDIS_WORK_ITEM WorkItem, 
	PVOID Context
	)
{
	PMP_ADAPTER 	Adapter = (PMP_ADAPTER)Context;

	UNREFERENCED_PARAMETER(WorkItem);
		   
	SMSC_TRACE0(DBG_TX,"--> NICPostSendsWaitingWorkItemCallBack\r\n");


	//
	// If another thread has this same workitem running already 
	// wait for it finish.
	//

	NdisWaitEvent(&Adapter->SendsWaitingWorkItemReady, 0);
	NdisResetEvent(&Adapter->SendsWaitingWorkItemReady);

	//
	// Clear the flag to let the WorkItem structure be reused for
	// scheduling another one. Clear it before processing to avoid 
	// race condition.
	//
	InterlockedExchange(&Adapter->IsSendsWaitingWorkItemQueued, FALSE);

	if (MP_IS_READY_AND_LINKED_UP(Adapter) && MP_TEST_FLAG(Adapter, fMP_POST_WRITES))
	{
		if ((!IsListEmpty(&Adapter->SendTcbPendingList)) && (Adapter->nTcbBusySend == 0))
		{
			NICProcessTCB(Adapter, NULL);
		}

		if (!IsListEmpty(&Adapter->SendPktWaitList))
		{
			MPSend(Adapter);
		}
	}

	//
	// If another thread has this same workitem ready to run again
	// but waiting for us to finish, signal it to go ahead.
	//

	NdisSetEvent(&Adapter->SendsWaitingWorkItemReady);
	MP_DEC_REF(Adapter);	
  
	SMSC_TRACE0(DBG_TX,"<-- NICPostSendsWaitingWorkItemCallBack\r\n");		  

}



/*++

Routine Description:

   Workitem to post all the waiting queued sends. This workitem is scheduled 
   from the SendsCompletionWorkItemCallBack if there're packets to send in the 
   send wait queue.
	  
Arguments:

	WorkItem - Pointer to workitem
	
	Context - pointer to Adapter.

Return Value:

	None.

--*/

VOID
NICPostSendsWaitingCallBack(
	PMP_ADAPTER 	Adapter
	)
{
		   
	SMSC_TRACE0(DBG_TX,"--> NICPostSendsWaitingWorkItemCallBack\r\n");

	if (MP_IS_READY_AND_LINKED_UP(Adapter) && MP_TEST_FLAG(Adapter, fMP_POST_WRITES))
	{
		if ((!IsListEmpty(&Adapter->SendTcbPendingList)) && (Adapter->nTcbBusySend == 0))
		{
			NICProcessTCB(Adapter, NULL);
		}

		if (!IsListEmpty(&Adapter->SendPktWaitList))
		{
			MPSend(Adapter);
		}
	}
  
	SMSC_TRACE0(DBG_TX,"<-- NICPostSendsWaitingWorkItemCallBack\r\n");		  

}


/*++

Routine Description:

		As a deserialized driver, we are responsible for holding incoming send 
		packets in our internal queue until they can be transmitted over the 
		network and for preserving the protocol-determined ordering of packet
		descriptors incoming to its MiniportSendPackets function. 

	We send as many as we can. Care is taken with locking so that if we run 
	out of resources (TCBs) the packet chain is put at the head of the queue (to 
	be sent sooner rather than later) while waiting for resources to free up.

		A deserialized miniport driver must complete each incoming send packet
		with NdisMSendComplete, and it cannot call NdisMSendResourcesAvailable. 

	
Arguments:

	PMP_ADAPTER 	Adapter

Return Value:

	None
	
--*/
VOID
MPSend(
PMP_ADAPTER 	Adapter)
{
	PLIST_ENTRY 	  pEntry = NULL;
	PTCB			  pTCB = NULL;
#ifdef BIDIMENSIONAL_SEND_LIST
	NDIS_STATUS 	  Status;
#endif
	NTSTATUS		  ntStatus;
	PNDIS_PACKET	  HeadPacket = NULL, Packet = NULL;
	BOOLEAN 		  BadPacket = FALSE, ChainEmpty = TRUE;
			
	SMSC_TRACE0(DBG_TX,"--> MPSend\r\n");
	
	// While we are copying packets into a TCB
	// no other thread can take stuff out of the 
	// queue so that we preserve queue ordering
	
	// Now try to send as many of the packets in the ready queue as possible.


	 #ifndef BIDIMENSIONAL_SEND_LIST

	// Now try to send as many of the packets in the ready queue as possible.

	while(TRUE)
	{			 
		// Get a packet
		NdisAcquireSpinLock(&Adapter->SendLock);
		pEntry = (PLIST_ENTRY)RemoveHeadList(&Adapter->SendPktWaitList);

		if(pEntry == &Adapter->SendPktWaitList)
		{
			// Nothing new to send, check if anything's pending to send
			NdisReleaseSpinLock(&Adapter->SendLock); 

			if (pTCB != NULL)
			{
				// We had a TCB pending and there are no more packets pending in the 
				// queue. So process the TCB (if no data return it, otherwise send it).
				// and get out of here right after.
			   SMSC_TRACE1(DBG_TX,"TCB 0x%p Pending, processing it...\r\n", pTCB);
				
				ntStatus = NICProcessTCB(Adapter, pTCB);
				if ( !NT_SUCCESS ( ntStatus ) ) 
				{
					SMSC_TRACE1(DBG_ERROR, "NICProcessTCB failed %x\r\n", ntStatus );
				} 

				// We will now break out of this function so we are releasing
				// the TCB, but since we are getting out, no need to formally 
				// set it to NULL, thus commenting them out.
				// pTCB = NULL;
			}
			break;
		}

		// OK got something from the queue, get the actual packet then.
		Packet = CONTAINING_RECORD(pEntry, NDIS_PACKET, MiniportReservedEx);
		if(Packet == NULL)
		{
			// This should NEVER happen !
 
			SMSC_TRACE0(DBG_ERROR, "Packet pointer in List is NULL\r\n");
			NdisReleaseSpinLock(&Adapter->SendLock); 
			continue;
		}

		// OK, we have a valid packet to send, let's get the TCB to put it in

		if (pTCB == NULL)
		{
			// A pending TCB has priority over the Free TCB
			pTCB = (PTCB) RemoveHeadList(&Adapter->SendTcbPendingList); 	   
			if(pTCB != (PTCB) &Adapter->SendTcbPendingList)
			{
				NdisInterlockedDecrement(&Adapter->nTcbPendingSend);
			}
			else
			{
				// No Pending TCB, Try to get a free TCB		
				pTCB = (PTCB) RemoveHeadList(&Adapter->SendTcbFreeList);		

			   
			}
		}
		

		if(pTCB == (PTCB) &Adapter->SendTcbFreeList)
		{			 
			//
			// Not able to get TCB for this send packet. So queue packet 
			// back for later transmission and get out.
			//

			 InsertHeadList( &Adapter->SendPktWaitList, 
							(PLIST_ENTRY)&Packet->MiniportReservedEx[0]); 

			NdisReleaseSpinLock(&Adapter->SendLock); 

			break;
		}
		else
		{	 
			// Make sure if we use RemoveEntryList on this TCB later we don't corrupt the queue.
			NdisInitializeListHead(&pTCB->List);
			
			NdisReleaseSpinLock(&Adapter->SendLock); 

			//
			// Copy the content of the packet to a the TCB's 
			// buffer. We need to do this because the target driver
			// is not capable of handling chained buffers.
			//

			if(!NICCopyPacket(Adapter, pTCB, Packet))
			{
				SMSC_TRACE1(DBG_TX, "Packet 0x%p is bad or was dropped due to filtering\r\n", Packet);
				
				NDIS_SET_PACKET_STATUS(Packet, NDIS_STATUS_FAILURE);
			}
			else 
			{
				// Got one packet in successfully, count it.
				pTCB->nPktCount++;

				NDIS_SET_PACKET_STATUS(Packet, NDIS_STATUS_SUCCESS);
			}

			// Let's see if this TCB is ready for sending now.
			if ((pTCB->bFull == TRUE) || (Adapter->TxOnePktperTcb))
			{
				// Process the TCB.

				ntStatus = NICProcessTCB(Adapter, pTCB);
				if ( !NT_SUCCESS ( ntStatus ) ) 
				{
					// TCB has already been put in the free list, so nothing else to do here.
					SMSC_TRACE1(DBG_ERROR, "NICProcessTCB failed %x\r\n", ntStatus );
				} 

				// Done with this TCB, signal it for the start of the while loop.
				pTCB = NULL;
			}

			// Return packet to Ndis in we are not in "delayed completion" return mode 
			if (!Adapter->DelayTxComplete)
			{
				Adapter->nSendPacketsCompleted++;
				
				NdisMSendComplete(
					Adapter->AdapterHandle,
					Packet,
					NDIS_GET_PACKET_STATUS(Packet));
			}
		}
	}

	#else

	
	while(TRUE)
	
	{ 
		
				
		if (HeadPacket == NULL)
		{
			// Get a Head Packet
			NdisAcquireSpinLock(&Adapter->SendLock);
			pEntry = (PLIST_ENTRY)RemoveHeadList(&Adapter->SendPktWaitList);

			if(pEntry == &Adapter->SendPktWaitList)
			{
				// Nothing new to send, check if anything's pending to send
				NdisReleaseSpinLock(&Adapter->SendLock); 

				if (pTCB != NULL)
				{
					// We had a TCB pending and there are no more packets pending in the 
					// queue. So process the TCB (if no data return it, otherwise send it).
					// and get out of here right after.
					SMSC_TRACE1(DBG_TX,"TCB 0x%p Pending, processing it...\r\n", pTCB);
					
					ntStatus = NICProcessTCB(Adapter, pTCB);

					if ( !NT_SUCCESS ( ntStatus ) ) {
						SMSC_TRACE1(DBG_ERROR,"NICProcessTCB failed 1  %x\r\n", ntStatus);
					} 
				
				}
				break;
			}

			// OK got something from the queue, get the actual packet then.
			HeadPacket = CONTAINING_RECORD(pEntry, NDIS_PACKET, MiniportReservedEx);
			if(HeadPacket == NULL)
			{
				SMSC_TRACE0(DBG_ERROR,"Packet pointer in List is NULL\r\n");
				NdisReleaseSpinLock(&Adapter->SendLock); 
				continue;
			}

			SMSC_TRACE1(DBG_TX,"Got Packet 0x%p from queue\r\n", HeadPacket);
			Packet = HeadPacket;
		}
		else
		{
			if (Packet == NULL)
			{
				SMSC_TRACE0(DBG_TX,"No Packet Pending\r\n");
				// Get a new entry from the chain
				ASSERT(HeadPacket);

				if ((pEntry = ((PLIST_ENTRY)&HeadPacket->Reserved[0])->Flink) == \
					(PLIST_ENTRY)&HeadPacket->Reserved[0])
				{
					ChainEmpty = TRUE;
				}
				else
				{
					ChainEmpty = FALSE;
				}

				Status = NDIS_GET_PACKET_STATUS(HeadPacket);
						
				// Lets complete the previous packet to NDIS if we need to, but be very careful 
				// with queuing so that we make sure not to complete them more than once.
				if ((!Adapter->DelayTxComplete) && ((Status == NDIS_STATUS_SUCCESS) || (Status == NDIS_STATUS_FAILURE)))
				{
					Adapter->nSendPacketsCompleted++;

					RemoveEntryList((PLIST_ENTRY)&HeadPacket->Reserved[0]);
					
					SMSC_TRACE2(DBG_TX,"Calling NdisMSendComplete %s on Packet 0x%p\r\n", 
							(Status == NDIS_STATUS_SUCCESS) ? "NDIS_STATUS_SUCCESS" : "NDIS_STATUS_FAILURE",
							HeadPacket );
					
					NdisMSendComplete(
						Adapter->AdapterHandle,
						HeadPacket,
						Status);
				}						

				// Done with the HeadPacket, mark it so.
				HeadPacket = NULL;					
						
				if (ChainEmpty)
				{
					// Nothing new to send in the chain, check if anything's pending to send
					if ((pTCB != NULL) && (Adapter->TxOnePktperTcb))
					{
						// Process the TCB.
						SMSC_TRACE1(DBG_TX,"TCB 0x%p Pending and one pkt per TCB so process it.\r\n", pTCB);
						ntStatus = NICProcessTCB(Adapter, pTCB);
						if ( !NT_SUCCESS ( ntStatus ) ) {
							SMSC_TRACE2(DBG_ERROR,"NICProcessTCB failed 2  %x, %d %d\n", ntStatus, MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED));
						} 
						// Done with the TCB, mark it so.
						pTCB = NULL;
					}

					// Move on to see if there are any more chains in the queue
					SMSC_TRACE0(DBG_TX,"No packets in chain, continue to next chain...\r\n");
					continue;
				}

				// OK got something from the chain, get the actual packet then.
				Packet = CONTAINING_RECORD(pEntry, NDIS_PACKET, Reserved);
				if(Packet == NULL)
				{
					SMSC_TRACE0(DBG_ERROR, "Packet pointer in List is NULL\r\n");
					continue;
				}

				if ((pTCB != NULL) && (Adapter->TxOnePktperTcb))
				{
					SMSC_TRACE1(DBG_TX,"TCB 0x%p Pending and one pkt per TCB so process it.\r\n", pTCB);

					ntStatus = NICProcessTCB(Adapter, pTCB);
					if ( !NT_SUCCESS ( ntStatus ) ) {
						SMSC_TRACE1(DBG_ERROR,"NICProcessTCB failed 3  %x\r\n", ntStatus);
					} 

					// Done with the TCB, mark it so.
					pTCB = NULL;
				}

				// The new packet is now the head packet
				HeadPacket = Packet;					
				SMSC_TRACE1(DBG_TX,"Packet 0x%p is now a new HeadPacket\r\n", Packet);
				NdisAcquireSpinLock(&Adapter->SendLock);
			}
			else
			{
				BadPacket = FALSE;

				SMSC_TRACE1(DBG_TX,"Found 0x%p Packet Pending\r\n", Packet);

				// This packet was pending from the previous "go around the while"
				// which means we could not send it before, either due to lack 
				// of space or because it's a bad packet. 

				if (pTCB == NULL)
				{
					SMSC_TRACE0(DBG_ERROR, "TCB pointer NULL with a pending packet\r\n");
				}
				else
				{
					// So process the current TCB (if no data return it, otherwise 
					// send it). 

					SMSC_TRACE1(DBG_TX,"Processing pending TCB 0x%p\r\n", pTCB);

					// Save the packet status for later first.
					if (!pTCB->ulSize)
						BadPacket = TRUE;

					ntStatus = NICProcessTCB(Adapter, pTCB);

					if ( !NT_SUCCESS ( ntStatus ) ) {
						SMSC_TRACE1(DBG_ERROR,"NICProcessTCB failed 4  %x\r\n", ntStatus);
					} 
					
					 
				}

				// If the packet was bad return it as failed and go to the next one.
				if (BadPacket)
				{
					SMSC_TRACE1(DBG_ERROR, "Packet 0x%p is bad and could not be copied, returning FAILURE status\r\n", Packet);
				
					NDIS_SET_PACKET_STATUS(Packet, NDIS_STATUS_FAILURE);

					HeadPacket = Packet;
				}
				else
				{
					// Now put this packet and what's left of its chain back in the queue
					NdisAcquireSpinLock(&Adapter->SendLock);

					InsertHeadList( &Adapter->SendPktWaitList, 
								(PLIST_ENTRY)&Packet->MiniportReservedEx[0]); 

					HeadPacket = NULL;

					NdisReleaseSpinLock(&Adapter->SendLock); 
				}

				// Now release the TCB and packet pointers
				pTCB = NULL;
				Packet = NULL;
				
				continue;
			}//endif if (packet==null)
		}//endof if (headpacket==null)

		//
		// Continue processing current TCB or get a free TCB block
		// At this point Packet == HeadPacket and it contains the 
		// packet we need to send now.
		//
		if (pTCB == NULL)
		{
			// A pending TCB has priority over the Free TCB
			pTCB = (PTCB) RemoveHeadList(&Adapter->SendTcbPendingList); 	   
			if(pTCB != (PTCB) &Adapter->SendTcbPendingList)
			{
				NdisInterlockedDecrement(&Adapter->nTcbPendingSend);
				SMSC_TRACE1(DBG_TX,"Pending TCB %p\r\n", pTCB);
			
			}
			else
			{
				// No Pending TCB, Try to get a free TCB		
				pTCB = (PTCB) RemoveHeadList(&Adapter->SendTcbFreeList);		

				if(pTCB != (PTCB) &Adapter->SendTcbFreeList)
				{
					SMSC_TRACE1(DBG_TX,"New TCB %p\r\n", pTCB);
					ASSERT(pTCB->ulSize == 0);
				}
				
			}
		}
		else
		{
			SMSC_TRACE1(DBG_TX,"In Use TCB %p\r\n", pTCB);
		}

		if(pTCB == (PTCB) &Adapter->SendTcbFreeList)
		{			 
			pTCB = NULL;

			//
			// Not able to get TCB block for this send packet. So queue packet chain
			// back for later transmission and get out.
			//

			SMSC_TRACE1(DBG_TX,"Can't allocate a TCB, returning Packet 0x%p (and it's chain if any) to the wait list!\r\n", Packet); 
			InsertHeadList( &Adapter->SendPktWaitList, 
							(PLIST_ENTRY)&Packet->MiniportReservedEx[0]); 
				
			NdisReleaseSpinLock(&Adapter->SendLock); 

			break;
		}
		else
		{	 
			// Make sure if we use RemoveEntryList on this TCB later we don't corrupt the queue.
			NdisInitializeListHead(&pTCB->List);
			
			NdisReleaseSpinLock(&Adapter->SendLock); 

			//
			// Copy the content of the packet to a the TCB's 
			// buffer. We need to do this because the target driver
			// is not capable of handling chained buffers.
			//

			if(!NICCopyPacket(Adapter, pTCB, Packet))
			{
				SMSC_TRACE2(DBG_TX,"No space in TCB 0x%p or bad packet 0x%p\r\n", pTCB, Packet);
				continue;
			}
			else 
			{
				// Got one packet in successfully, count it.
				pTCB->nPktCount++;

				NDIS_SET_PACKET_STATUS(Packet, NDIS_STATUS_SUCCESS);

				// Since we need the HeadPacket to retrieve the next packet of the chain we 
				// cannot NdisMSendComplete here, we'll do it when we come back into the 
				// while loop above

				// Packet was copied successfully, mark it by deleting the pointer reference
				Packet = NULL;
			}
		}
	}
	#endif
	SMSC_TRACE0(DBG_TX,"<-- MPSend\r\n");

}

/*++

Routine Description:

		Send Packet Array handler. Called by NDIS whenever a protocol
		bound to our miniport sends one or more packets.

		The input packet descriptor pointers have been ordered according 
		to the order in which the packets should be sent over the network 
		by the protocol driver that set up the packet array. The NDIS 
		library preserves the protocol-determined ordering when it submits
		each packet array to MiniportSendPackets

	A deserialized miniport driver must complete each incoming send packet
		with NdisMSendComplete, and it cannot call NdisMSendResourcesAvailable. 

	This is the main entry point which just puts the packets in the send 
	wait queue and then we call the serializing MPSend function which does 
	all actual transmission handling and completes the packets when 
	done. The only case in which this routine completes packets itself is
	when the device was not ready for transmissions (either no link or
	transmissions specifically disabled via !fMP_POST_WRITES)
	
Arguments:

	MiniportAdapterContext	  Pointer to our adapter context
	PacketArray 			  Set of packets to send
	NumberOfPackets 		  Length of above array

Return Value:

	None
	
--*/
VOID 
MPSendPackets(
	IN	NDIS_HANDLE 			MiniportAdapterContext,
	IN	PPNDIS_PACKET			PacketArray,
	IN	UINT					NumberOfPackets)
{
	PMP_ADAPTER 	  Adapter;
	UINT			  PacketCount, CompletedCount = 0;
	PNDIS_PACKET	  Packet, HeadPacket = NULL;


	SMSC_TRACE1(DBG_TX,"--> MPSendPackets (%d packets)\r\n", NumberOfPackets);


	Adapter = (PMP_ADAPTER)MiniportAdapterContext;
   
	#ifndef BIDIMENSIONAL_SEND_LIST


	if(!(MP_IS_READY_AND_LINKED_UP(Adapter) && MP_TEST_FLAG(Adapter, fMP_POST_WRITES)))
	{
		for(PacketCount=0; PacketCount < NumberOfPackets; PacketCount++)
		{
			// Grab packet from array
			Packet = PacketArray[PacketCount];
		
			// Check for a zero pointer, should not happen but is recoverable
			if(Packet == NULL)
			{

				SMSC_TRACE1(DBG_ERROR,"Packet pointer #%d in Packet Array is NULL\r\n", PacketCount);
				continue;
			}

			Adapter->nSendPacketsArrived++;

			NDIS_SET_PACKET_STATUS(Packet, NDIS_STATUS_FAILURE);

			NdisMSendComplete(
				Adapter->AdapterHandle,
				Packet,
				NDIS_STATUS_FAILURE);

			Adapter->nSendPacketsCompleted++;
		}
	}
	else
	{
		NdisAcquireSpinLock(&Adapter->SendLock);
		for(PacketCount=0; PacketCount < NumberOfPackets; PacketCount++)
		{
			// Grab packet from array
			Packet = PacketArray[PacketCount];
		
			// Check for a zero pointer, should not happen but is recoverable
			if(Packet == NULL)
			{
 
				SMSC_TRACE1(DBG_ERROR,"Packet pointer #%d in Packet Array is NULL\r\n", PacketCount);
				continue;
			}

			Adapter->nSendPacketsArrived++;

			InsertTailList( &Adapter->SendPktWaitList, 
							(PLIST_ENTRY)&Packet->MiniportReservedEx[0]); 

			Adapter->nSendPacketsPended++;
			NDIS_SET_PACKET_STATUS(Packet, NDIS_STATUS_PENDING);
		}
		NdisReleaseSpinLock(&Adapter->SendLock); 
	}

	// Now Send as many packets we have queued as we can.
	if (!IsListEmpty(&Adapter->SendPktWaitList) && MP_IS_READY_AND_LINKED_UP(Adapter))
	{
		   
		 NICPostSendsWaitingCallBack(Adapter);

	 }
	

	#else

	

	for(PacketCount=0; PacketCount < NumberOfPackets; PacketCount++)
	{
		// Grab packet from array
		Packet = PacketArray[PacketCount];
		
		// Check for a zero pointer, should not happen but is recoverable
		if(Packet == NULL)
		{
			SMSC_TRACE1(DBG_ERROR,"Packet pointer #%d in Packet Array is NULL\r\n", PacketCount);
			continue;
		}

		SMSC_TRACE2(DBG_TX,"Adapter %p, Packet %p\r\n", Adapter, Packet);
		
		Adapter->nSendPacketsArrived++;
		
		if(MP_IS_READY_AND_LINKED_UP(Adapter) && MP_TEST_FLAG(Adapter, fMP_POST_WRITES))
		{
			if (!HeadPacket)
			{
				// Initialize this packet as the Head of it's chain
				SMSC_TRACE1(DBG_TX,"Packet 0x%p designated head of chain\r\n", Packet);
				NdisInitializeListHead((PLIST_ENTRY)&Packet->Reserved[0]);
				HeadPacket = Packet;
			}
			else
			{
				SMSC_TRACE2(DBG_TX,"Inserting Packet 0x%p into chain of HeadPacket 0x%p\r\n", Packet, HeadPacket);
				NdisInterlockedInsertTailList(
					(PLIST_ENTRY)&HeadPacket->Reserved[0], 
					(PLIST_ENTRY)&Packet->Reserved[0], 
					&Adapter->SendLock);
			}

			Adapter->nSendPacketsPended++;
			NDIS_SET_PACKET_STATUS(Packet, NDIS_STATUS_PENDING);
		}
		else
		{
			SMSC_TRACE1(DBG_TX,"Calling NdisMSendComplete NDIS_STATUS_FAILURE on Packet 0x%p\r\n", Packet);

			NDIS_SET_PACKET_STATUS(Packet, NDIS_STATUS_FAILURE);

			CompletedCount++;
			
			NdisMSendComplete(
				Adapter->AdapterHandle,
				Packet,
				NDIS_STATUS_FAILURE);
		}
	}

	Adapter->nSendPacketsCompleted += CompletedCount;

	if (!HeadPacket)
	{
		if (CompletedCount != NumberOfPackets)
		{
			
			SMSC_TRACE0(DBG_ERROR,"No NON NULL packets in Packet array while posting writes and link up !!!\r\n");
			goto Exit_MPSendPackets;
		}
	}
	else
	{
		SMSC_TRACE1(DBG_TX,"Inserting Packet 0x%p and it's chain (if any) in Send Wait list\r\n", Packet);

		// Put the packet chain in the Send Wait List
		NdisInterlockedInsertTailList(
			&Adapter->SendPktWaitList, 
			(PLIST_ENTRY)&HeadPacket->MiniportReservedEx[0], 
			&Adapter->SendLock);

		 // Now Send as many packets we have queued as we can.
		if (!IsListEmpty(&Adapter->SendPktWaitList) && MP_IS_READY_AND_LINKED_UP(Adapter))
		{
		   
			   NICPostSendsWaitingCallBack(Adapter);
		}

			
	}
	Exit_MPSendPackets:
#endif

	SMSC_TRACE0(DBG_TX,"<-- MPSendPackets\r\n");

	return;
}


/*++

Routine Description:

		This routine Issue a Bulk Out request  to the target device to send
		the network packet out to the device.
		
Arguments:

		Adapter    - pointer to the MP_ADAPTER structure
		pTCB		 -	Pointer to the TCB block


Return Value:

		NT status code

--*/

NTSTATUS
NICPostWriteRequest(
	PMP_ADAPTER Adapter,
	PTCB	pTCB
	)

{
			  
	SMSC_TRACE0(DBG_TX,"--> NICPostWriteRequest\r\n");

 
	if (!ACCEPT_IO(Adapter)) {
		SMSC_TRACE0(DBG_ERROR,"NICPostWriteRequest(),Device is not ready for I/O request!!!\r\n");	
		return STATUS_UNSUCCESSFUL;
	}

	ResetEvent(pTCB->SyncEvent);

	pTCB->dwErr = IssueBulkTransfer( Adapter->UsbFuncs,
									   Adapter->BulkOut.hPipe,
									   NICWriteRequestCompletion,
									   pTCB,
									   (USB_OUT_TRANSFER|USB_SHORT_TRANSFER_OK),
									   pTCB->pData, 0,
									   pTCB->ulSize,
									  &pTCB->dwTransferLength,
									   0,
									  &pTCB->dwUsbErr );


	SetEvent(pTCB->SyncEvent);

//	ASSERT(!(!pTCB->dwErr && !pTCB->dwUsbErr));

	if ( (ERROR_SUCCESS ==	pTCB->dwErr) && (USB_NO_ERROR ==  pTCB->dwUsbErr)  ) 
	{
		SMSC_TRACE2(DBG_ERROR,"Write Request, NULL Transfer, pTCB->dwErr:%d  pTCB->dwUsbErr:0x%x)\r\n",  pTCB->dwErr,	pTCB->dwUsbErr);
		return STATUS_UNSUCCESSFUL;
	}
		
	
	if (STATUS_SUCCESS != pTCB->dwErr) {
		SMSC_TRACE0(DBG_ERROR,"Issue Bulk out Transfer ERROR\r\n");
		return STATUS_UNSUCCESSFUL;
			
	} 
		
	SMSC_TRACE0(DBG_TX,"<-- NICPostWriteRequest\r\n");

	return STATUS_SUCCESS;
}

/*++

Routine Description:

	This routine Stops the Tx Path at both Mac and USB and then flushes 
	the Tx Fifo pointers.

Arguments:

	Adapter - pointer to our adapter context structure

Return Value:
	NIC_STATUS_XXX
	
--*/
NIC_STATUS
NICStopAndFlushTxPath(
	PMP_ADAPTER Adapter
	)
{

	DWORD dwTemp;
	NIC_STATUS Status = NIC_STATUS_FAILURE;
	SMSC_TRACE1(DBG_TX,"--> NICStopAndFlushTxPath,Adapter=%p\r\n",Adapter);

	// Since it can be called from MP_Halt when MP_Initialize fails
	if(!MP_TEST_FLAG(Adapter, fMP_SEND_SIDE_RESOURCE_ALLOCATED)){
		goto Exit_NICStopAndFlushTxPath;
	}

	Status = NICStopTxPath(Adapter);
	dwTemp= ETH_TX_FIFO_FLUSH_;
	Status |= NICWriteRegister(Adapter, ETH_TX_CFG, &dwTemp);

	// Should self clear way before the read.
	Status |= NICReadRegister(Adapter, ETH_TX_CFG, &dwTemp);

	if ((Status == NIC_STATUS_SUCCESS) && (dwTemp & ETH_TX_FIFO_FLUSH_))
	{
		// Flush did not self clear!
		Status = NIC_STATUS_FAILURE;
	}

Exit_NICStopAndFlushTxPath:
   SMSC_TRACE0(DBG_TX,"<-- NICStopAndFlushTxPath\r\n");
   return Status;
}

/*++

Routine Description:

	This routine Stops the Tx Path at both Mac and USB

Arguments:

	Adapter - pointer to our adapter context structure

Return Value:
	NIC_STATUS_XXX
	
--*/
NIC_STATUS
NICStopTxPath(
	PMP_ADAPTER Adapter
	)
{ 
	DWORD Value32,dwTemp;
	NIC_STATUS Status;
	int Count=0;

	SMSC_TRACE1(DBG_TX,"--> NICStopTxPath,Adapter=%p\r\n",Adapter);
	if(!MP_TEST_FLAG(Adapter, fMP_SEND_SIDE_RESOURCE_ALLOCATED)){
		goto Exit_NICStopTxPath;
	}
	Status = NICStopBulkOut(Adapter);
	
	// Stop the Transmit path at SCSRs
	dwTemp=ETH_TX_STOP_;
	Status |= NICWriteRegister(Adapter, ETH_TX_CFG, &dwTemp);
	// The bit should self clear as soon as the packet makes it out of the Mac,
	// worst case is 10 Mbps HD and frame deferred to the maximum. This will 
	// be ~100ms tops. Assuming one register read per (micro)frame the case of 
	// high speed USB - 125us register read cycle time - is the worse and 
	// would need up to 800 reads. Let's just round up to 1000.
	do 
	{
		Status |= NICReadRegister(Adapter, ETH_TX_CFG, &Value32);
		// Let it try to do the 1000 reads even if the reg reads are failing
		// If the previous write did go thru at least this way we have a better 
		// chance of making sure the transmit path did stop.
	}
	while ( (++Count<1000) && ((Value32 & (ETH_TX_STOP_ | ETH_TX_ON_)) != 0) );

	// Disable Mac TX 
	Status |= NICReadRegister(Adapter, MAC_CR, &Value32);
	dwTemp=Value32 & (~MAC_CR_TXEN_);
	Status |= NICWriteRegister(Adapter, MAC_CR, &dwTemp);
Exit_NICStopTxPath:
	SMSC_TRACE0(DBG_TX,"<-- NICStopTxPath\r\n");
	return Status;
}

/*++

Routine Description:

		This routine prevents further bulk out submissions and cancels 
		any outstanding write requests waiting for their completion.

Arguments:

		Adapter - pointer to our adapter context structure

Return Value:
	NIC_STATUS_XXX
	
--*/
NIC_STATUS
NICStopBulkOut(
	PMP_ADAPTER Adapter
	)
{ 
	BOOLEAN 		  bDone=TRUE;
	LONG			  nWaitCount = 0;
	NIC_STATUS Status = NIC_STATUS_SUCCESS;

	SMSC_TRACE1(DBG_SHUT_DOWN,"--> NICStopBulkOut,Adapter=%p\r\n",Adapter);

	if(!MP_TEST_FLAG(Adapter, fMP_SEND_SIDE_RESOURCE_ALLOCATED)){
			goto Exit_NICStopBulkOut;
		}
	if (Adapter->IsSendReady)
		NdisWaitEvent(&Adapter->SendsWaitingWorkItemReady, 0);

	MP_CLEAR_FLAG(Adapter, fMP_POST_WRITES);
	
	if (!MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED)) {

		Adapter->BulkOut.PipeTransferAborted=TRUE;
		if ( !( Adapter->UsbFuncs->lpAbortPipeTransfers(Adapter->BulkOut.hPipe,USB_NO_WAIT)))
		{
			SMSC_TRACE0(DBG_ERROR,"Bulk Out Pipe Abort failed !!!\r\n");
			
		}
	}
	
	NICFreePendingSendPackets(Adapter);

	// Wait for all bulk out completions.
	while(TRUE)
	{
		bDone = TRUE;

		//
		// Are all the bulk out "in process" and "pending" transactions completed ?
		//
		NdisAcquireSpinLock(&Adapter->SendLock);
		if(Adapter->nTcbBusySend + Adapter->nTcbPendingSend +Adapter->nTcbCompletionSend)
		{
				   
			SMSC_TRACE3(DBG_TX,"nTcbBusySend = %d, nTcbPendingSend = %d, nTcbCompletionSend= %d\r\n", 
						Adapter->nTcbBusySend,
						Adapter->nTcbPendingSend,
						Adapter->nTcbCompletionSend);
			bDone = FALSE;
		}
		NdisReleaseSpinLock(&Adapter->SendLock);
		
		if(bDone)
		{
			break;	 
		}

		if((++nWaitCount % NIC_1_SEC_TIMEOUT) == 0)
		{
			SMSC_TRACE1(DBG_ERROR,"Bulk Out Cancel timed out after %d second!!! COMMITING LEAK !!!\r\n", NIC_1_SEC_TIMEOUT);
			SMSC_TRACE1(DBG_ERROR,"   nTcbBusySend = %d\r\n", Adapter->nTcbBusySend);
			SMSC_TRACE1(DBG_ERROR,"   nTcbPendingSend = %d\r\n", Adapter->nTcbPendingSend);
			SMSC_TRACE1(DBG_ERROR,"   nTcbCompletionSend = %d\r\n", Adapter->nTcbCompletionSend);
			Status = NIC_STATUS_FAILURE;
			ASSERT(FALSE);		 
			break;
		}

		NdisMSleep(NDIS_SLEEP_1_MS);		
	}
Exit_NICStopBulkOut:
	SMSC_TRACE0(DBG_SHUT_DOWN,"<-- NICStopBulkOut\r\n");
	return Status;
}

/*++

Routine Description:

	This routine reinitializes the TCB block and puts it back
		into the SendTcbFreeList from SendTcbBusyList for reuse. 

	
Arguments:

		Adapter    - pointer to the adapter structure
		pTCB	  - pointer to TCB block
		
Return Value:

		VOID

--*/

VOID 
NICFreeSendTCB(
	IN PMP_ADAPTER Adapter,
	IN PTCB pTCB)
{
	SMSC_TRACE1(DBG_TX,"--> NICFreeSendTCB %p\r\n", pTCB);
	
	pTCB->OrgSendPacket = NULL;
	pTCB->bFull = FALSE;
	pTCB->ulSize = 0;
	pTCB->nPktCount = 0;
	  
	ASSERT(!pTCB->Ref);

	//
	// Insert the TCB back in the send free list	 
	//
	NdisAcquireSpinLock(&Adapter->SendLock);
	
	RemoveEntryList(&pTCB->List);
	NdisInterlockedDecrement(&Adapter->nTcbBusySend);
	ASSERT(Adapter->nTcbBusySend >= 0);
	
	InsertTailList(&Adapter->SendTcbFreeList, &pTCB->List);

	NdisReleaseSpinLock(&Adapter->SendLock); 

	SMSC_TRACE0(DBG_TX,"<-- NICFreeSendTCB\r\n");
}


/*++

Routine Description:

	This routine reinitializes the TCB block and puts it back
		into the SendTcbFreeList from SendTcbcompletionList for reuse. 
		
Arguments:

		Adapter    - pointer to the adapter structure
		pTCB	  - pointer to TCB block

Return Value:

		VOID

--*/
VOID 
NICFreeCompletionTCB(
	IN PMP_ADAPTER Adapter,
	IN PTCB pTCB)
{
	SMSC_TRACE1(DBG_TX,"--> NICFreeCompletionTCB %p\r\n", pTCB);
	
	pTCB->OrgSendPacket = NULL;
	pTCB->bFull = FALSE;
	pTCB->ulSize = 0;
	pTCB->nPktCount = 0;
	
	   
	ASSERT(!pTCB->Ref);   
	//
	// Insert the TCB back in the send free list	 
	//
	NdisAcquireSpinLock(&Adapter->SendLock);
	   
	InsertTailList(&Adapter->SendTcbFreeList, &pTCB->List);

	NdisReleaseSpinLock(&Adapter->SendLock); 

	SMSC_TRACE0(DBG_TX,"<-- NICFreeCompletionTCB\r\n");
}


/*++

Routine Description:

	This routine is called by the Halt handler to fail all
	the queued up SendPackets because the device is either
	gone, being stopped for resource rebalance.
	
Arguments:

	Adapter    - pointer to the adapter structure

Return Value:

	VOID

--*/
VOID 
NICFreeQueuedSendPackets(
	PMP_ADAPTER Adapter
	)
{
	PLIST_ENTRY 	  pEntry = NULL;
	PNDIS_PACKET	  HeadPacket = NULL, TailPacket = NULL, Packet = NULL;

	SMSC_TRACE0(DBG_TX,"--> NICFreeQueuedSendPackets\r\n");

	#ifndef BIDIMENSIONAL_SEND_LIST

	 while(MP_TEST_FLAG(Adapter, fMP_SEND_SIDE_RESOURCE_ALLOCATED))
	{
		// Get a packet out of the Wait list

		pEntry = (PLIST_ENTRY) NdisInterlockedRemoveHeadList(
						&Adapter->SendPktWaitList, 
						&Adapter->SendLock);
		if(!pEntry)
		{
			break;
		}
		
		Packet = CONTAINING_RECORD(pEntry, NDIS_PACKET, MiniportReservedEx);

 
		if (Packet != NULL)
		{
		 
			Adapter->nSendPacketsCompleted++;

			NdisMSendComplete(
				Adapter->AdapterHandle,
				Packet,
				NDIS_STATUS_REQUEST_ABORTED);
		}
	}

	#else
 
	while(MP_TEST_FLAG(Adapter, fMP_SEND_SIDE_RESOURCE_ALLOCATED))
	{
		// Get a chain out of the Wait list
		pEntry = (PLIST_ENTRY) NdisInterlockedRemoveHeadList(
						&Adapter->SendPktWaitList, 
						&Adapter->SendLock);
		if(!pEntry)
		{
			break;
		}

		// In order to preserve the order on completions, use as chain head the 
		// the tail packet, so that the while loop below takes the packets to be 
		// completed out of the chain in the proper order (head, ..., tail)
		HeadPacket = CONTAINING_RECORD(pEntry, NDIS_PACKET, MiniportReservedEx);

		pEntry = (PLIST_ENTRY) &HeadPacket->Reserved[0];
		TailPacket = CONTAINING_RECORD(pEntry->Blink, NDIS_PACKET, Reserved);

		ASSERT(TailPacket);

		pEntry = (PLIST_ENTRY)RemoveHeadList((PLIST_ENTRY)&TailPacket->Reserved[0]);

		while(pEntry != (PLIST_ENTRY)&TailPacket->Reserved[0])
				
		{
			
			Packet = CONTAINING_RECORD(pEntry, NDIS_PACKET, Reserved);

			ASSERT(Packet);
			SMSC_TRACE1(DBG_TX,"Completing Packet 0x%p\r\n", Packet);

			Adapter->nSendPacketsCompleted++;
			
			NdisMSendComplete(
				Adapter->AdapterHandle,
				Packet,
				NDIS_STATUS_REQUEST_ABORTED);
			
			pEntry = (PLIST_ENTRY)RemoveHeadList((PLIST_ENTRY)&TailPacket->Reserved[0]);
			
		}
	
		SMSC_TRACE1(DBG_TX,"Completing Packet 0x%p\r\n", Packet);

		Adapter->nSendPacketsCompleted++;
		
		NdisMSendComplete(
			Adapter->AdapterHandle,
			TailPacket,
			NDIS_STATUS_REQUEST_ABORTED);
	}
	#endif

	SMSC_TRACE0(DBG_TX,"<-- NICFreeQueuedSendPackets\r\n");

}


/*++

Routine Description:

		This routine copies the packet data into the TCB data block.
			
Arguments:

		Adapter   - pointer to the MP_ADAPTER structure
		pTCB	  - pointer to TCB block
		Packet	  - packet to be transfered.


Return Value:

		TRUE	if successfully copied the packet into the TCB
	FALSE	if could not copy the packet into the TCB. In this latter case:
				- if ulSize == 0 the packet was bad (failed virtual mem mapping)
				- if ulSize != 0 there was no space or the packet was bad

	Using back to back calls with the same packet, if it fails both we now for 
	sure the packet is bad.

--*/
BOOLEAN
NICCopyPacket(
	PMP_ADAPTER Adapter,
	PTCB pTCB, 
	PNDIS_PACKET Packet)
{
	PNDIS_BUFFER   CurrentBuffer = NULL;
	PVOID		   VirtualAddress = NULL;
	UINT		   CurrentLength;
	UINT		   BytesToCopy;
	UINT		   BytesCopied = 0;
	UINT		   PacketLength;	  
	PUCHAR		   pDest;
	BOOLEAN 	   bResult = TRUE;	
	UINT		   DestBufferSize = Adapter->TxBurstMax;		
	UINT		   SaveInitialTcbDataSize = pTCB->ulSize;
	ULONG		   AlignCount = 0;
		
	SMSC_TRACE0(DBG_TX,"--> NICCopyPacket\r\n");

	//
	// Query the packet to get length and pointer to the first buffer.
	//	  
	NdisQueryPacket(Packet,
					NULL,
					NULL,
					&CurrentBuffer,
					&PacketLength);

	
	ASSERT (pTCB->ulSize <= Adapter->TxBurstMax);

	if (pTCB->ulSize == 0)
	{
		//
		// This is a fresh/empty  TCB, initialize it 
		//
		pTCB->OrgSendPacket = Packet;
		pTCB->Ref = 1;
		NdisInitializeListHead(&pTCB->List);
	}
	else
	{
		/*
			*  We need to make sure TxA is dword aligned with respect to the 
			*  previous data we sent into the device (Hw requirement). Since 
			*  each BULK Out TCB has an integral number of packets, each TCB always 
			*  starts with TxA. Therefore if we make sure pTCB->ulSize is a 
			*  multiple of 4 before writing TxA we guarantee TxA's dword 
			*  alignment.
			*/

		AlignCount = (4 - (pTCB->ulSize % 4)) % 4;	// Get number of bytes to align ulSize to dword

		// Make sure there's space to fit: dword alignment padding + TxA + TxB + packet +4(VLan tag addition)
		if ( (Adapter->TxBurstMax - pTCB->ulSize) < ( AlignCount + 8 + PacketLength+4))
		{
			/*
					*  There is no space in this TCB for this packet, so
					*  return FALSE to the caller to signal it.
					*/
			SMSC_TRACE2(DBG_TX,"No space in TCB 0x%p for packet 0x%p\r\n", pTCB, Packet);

			bResult = FALSE;
			pTCB->bFull=TRUE;
			goto Exit_NICCopyPacket;
		}
	}

	/*
	 *	Accepted the packet !
	 */ 

	pTCB->ulSize += AlignCount; 				// Increase ulsize accordingly
	DestBufferSize -= AlignCount;				// Reduce avail space accordingly

	SMSC_TRACE2(DBG_TX,"Alignment insertion = %d into TCB %p\r\n", AlignCount, pTCB);

	pDest = (pTCB->pData + pTCB->ulSize);	// Setup the location to copy to.
	BytesToCopy = PacketLength;

	/*
	 *	VLAN TAG Insertion Checking Begins
	 */
	if (Adapter->PriorityVlanTag != 0)
	{
		NDIS_PACKET_8021Q_INFO VlanPriInfo;
		UINT32 UserPriority;
		UINT32 VlanID;
		BOOLEAN bInsertTag = FALSE;

		pDest+= 8UL;
		pTCB->ulSize += 8UL;
		DestBufferSize -=8UL;

		VlanPriInfo.Value = NDIS_PER_PACKET_INFO_FROM_PACKET(Packet, Ieee8021QInfo);
		UserPriority = VlanPriInfo.TagHeader.UserPriority;
		VlanID = VlanPriInfo.TagHeader.VlanId;
		SMSC_TRACE3(DBG_TX,"Tag. Value = %p, VlanID = %d, UserPriority = %d\r\n", VlanPriInfo.Value, VlanID, UserPriority);

		NdisQueryBufferSafe(
			CurrentBuffer,
			&VirtualAddress,
			&CurrentLength,
			NormalPagePriority);

		if(!VirtualAddress) 
		{
			SMSC_TRACE1(DBG_ERROR,"Could not get virtual address for Ndis buffer %p\r\n", CurrentBuffer);
			ASSERT(VirtualAddress != NULL);
			// This is an unrecoverable error for this packet, throw it away but signal 
			// caller to send out the previous ones we had already copied in this TCB.
			pTCB->ulSize = SaveInitialTcbDataSize;
			bResult = FALSE;
			goto Exit_NICCopyPacket;
		}

		if ((CurrentLength < ETH_HEADER_SIZE) || ((CurrentLength + 4) > DestBufferSize))
		{
			SMSC_TRACE4(DBG_ERROR,"First fragment for Ndis packet %p is smaller than Ethernet header; or dest buffer does not have enough space! Currentbuffer=%p, CurrentbufferLength = %d, DestBufferLength = %d\r\n", Packet, CurrentBuffer, CurrentLength, DestBufferSize);
			ASSERT(CurrentLength >= ETH_HEADER_SIZE);
			ASSERT((CurrentLength + 4) < DestBufferSize);
			pTCB->ulSize = SaveInitialTcbDataSize;
			bResult = FALSE;
			goto Exit_NICCopyPacket;
		}

		// Check send discard and generate tag heuristics.
		if ( *((PWORD)((PUCHAR)VirtualAddress + (2 * ETH_ADDR_SIZE))) == 0x0081 )
		{
			// Ethertype is 8021Q

			DWORD VlanTagInPkt;

			// Make sure we don't go past the end of the buffer to access the VLAN Tag
			if (CurrentLength < (ETH_HEADER_SIZE + 2))
			{
				SMSC_TRACE0(DBG_ERROR,"VLAN protocol but no space for Vlan Tag in first fragment!\r\n");
				ASSERT(CurrentLength < (ETH_HEADER_SIZE + 2));
				pTCB->ulSize = SaveInitialTcbDataSize;
				bResult = FALSE;
				goto Exit_NICCopyPacket;
			}

			// We have a VLAN Tag in the data portion of the packet.

			VlanTagInPkt = *((PDWORD)((PUCHAR)VirtualAddress + (2 * ETH_ADDR_SIZE)));

			if ( (VlanTagInPkt & 0xFF0F0000UL) == 0)
			{
				// VLAN Tag in data portion of packet exists and is ZERO

				if ( (UserPriority | VlanID | Adapter->VlanID) != 0 )
				{
					if ( 
						((VlanID & Adapter->VlanID) != 0) && 
						(VlanID != Adapter->VlanID)
					   )
					{
						// OOB and NIC VLAN ID deffer, drop
						SMSC_TRACE2(DBG_ERROR,"Fail the send because of VLAN ID mismatch in NIC VLAN (%d) and OOB data (%d)!\r\n", 
							 VlanID, Adapter->VlanID);
						pTCB->ulSize = SaveInitialTcbDataSize;
						bResult = FALSE;
						goto Exit_NICCopyPacket;
					}
					else
					{
						// Replace Zero tag with OOB Values and send.
						*((PDWORD)((PUCHAR)VirtualAddress + (2 * ETH_ADDR_SIZE))) = 
							((UserPriority << 21) | ((VlanID & 0xff) << 24) | ((VlanID & 0x0f00) << 8) | 0x0081);
						SMSC_TRACE1(DBG_TX,"VLAN Tag replaced. New tag = 0x%08X\n", *((PDWORD)((PUCHAR)VirtualAddress + (2 * ETH_ADDR_SIZE))));
					}
				}
				
			}
			else
			{
				// VLAN Tag in data portion of packet exists and is NON-ZERO
				if (Adapter->VlanID == 0)
				{
					// NIC VLAN is zero
					if ((VlanID | UserPriority) != 0)
					{
						// Both VLAN Tag in the data portion and OOB provided, invalid ; drop send
						SMSC_TRACE3(DBG_ERROR,"Both OOB (Vlan ID = %d, UserPriority = %d) and packet data Vlan tag (0x%08lx) exists! Dropping send packet\r\n", 
													VlanID, UserPriority, VlanTagInPkt);
						pTCB->ulSize = SaveInitialTcbDataSize;
						bResult = FALSE;
						goto Exit_NICCopyPacket;
					}
					
				}
				else
				{
					// NIC VLAN is NOT zero
					if (
						 ( Adapter->VlanID ==((VlanTagInPkt >> 24) | ((VlanTagInPkt & 0x000F0000) >> 8)) ) &&
						 ((VlanID & UserPriority) == 0)
					   )
					{
						// No Change, Send.
					}
					else
					{
						// Both VLAN Tag in the data portion and OOB provided; or missmatch; invalid ; drop send
						SMSC_TRACE4(DBG_ERROR,"Incompatible OOB (Vlan ID = %d, UserPriority = %d), packet data Vlan tag (0x%08lx), and NIC Vlan (%d)! Dropping send packet\r\n", 
													VlanID, UserPriority, VlanTagInPkt, Adapter->VlanID);
						pTCB->ulSize = SaveInitialTcbDataSize;
						bResult = FALSE;
						goto Exit_NICCopyPacket;
					}
				}
			}
		}
		else
		{
			// Ethertype is NOT 8021Q, there is no VLAN tag present in the data portion of the packet.
			if ((Adapter->VlanID | VlanID | UserPriority) == 0) 
			{
				// No Change, Send.
			}
			else
			{
				if ( (Adapter->VlanID != VlanID) && ((Adapter->VlanID & VlanID) != 0) )
				{
					SMSC_TRACE2(DBG_ERROR,"NIC Vlan ID (%d) mismatches OOB Vlan ID (%d)! Dropping send packet\r\n", 
						 Adapter->VlanID, VlanID);
					pTCB->ulSize = SaveInitialTcbDataSize;
					bResult = FALSE;
					goto Exit_NICCopyPacket;
				}
				else
				{
					// Need to insert a Vlan Tag, unless it's 0 use the OOB's
					bInsertTag = TRUE;
					if ((VlanID & UserPriority) == 0)
					{
						VlanID = Adapter->VlanID;
					}
				}
			}
		}

		if (bInsertTag)
		{
			PacketLength +=4;

			/*
					*  Copy the packet !
					*/ 

			SMSC_TRACE4(DBG_TX,"Copy pkt %p (size = %d) into TCB %p at offset %d\r\n", Packet, PacketLength, pTCB, pTCB->ulSize - 8UL);

			// Set the Cmd A and B

			*((DWORD *)(pDest-8UL)) = TX_CMD_A_FIRST_SEG_ | TX_CMD_A_LAST_SEG_ | PacketLength;
			*((DWORD *)(pDest-4UL)) = PacketLength;

			// Copy the SRC and DST Mac addresses.
			NdisMoveMemory(pDest, VirtualAddress, (2 * ETH_ADDR_SIZE));
			BytesCopied += (2 * ETH_ADDR_SIZE);
			DestBufferSize -= (2 * ETH_ADDR_SIZE);			  
			pDest += (2 * ETH_ADDR_SIZE);
			CurrentLength -= (2 * ETH_ADDR_SIZE);

			// Insert the Vlan Tag
			BytesToCopy += 4;		// Add the size of the Vlan tag 
			*((PUINT32)pDest) = ((UserPriority << 21) | ((VlanID & 0xff) << 24) | ((VlanID & 0x0f00) << 8) | 0x0081);
			SMSC_TRACE1(DBG_TX,"VLAN Tag inserted = 0x%08X\r\n", *((PUINT32)pDest));
			BytesCopied += 4;
			DestBufferSize -= 4;			
			pDest += 4;

			// Copy the rest of the first buffer's data.
			NdisMoveMemory(pDest, (PUCHAR)VirtualAddress + (2 * ETH_ADDR_SIZE), CurrentLength);
			BytesCopied += CurrentLength;
			DestBufferSize -= CurrentLength;			
			pDest += CurrentLength;

			NdisGetNextBuffer(
				CurrentBuffer,
				&CurrentBuffer);
		}
		else
		{
			/*
					*  Copy the packet !
					*/ 

			SMSC_TRACE4(DBG_TX,"Copy pkt %p (size = %d) into TCB %p at offset %d\n", Packet, PacketLength, pTCB, pTCB->ulSize - 8UL);

			// Set the Cmd A and B

			*((DWORD *)(pDest-8UL)) = TX_CMD_A_FIRST_SEG_ | TX_CMD_A_LAST_SEG_ | PacketLength;
			*((DWORD *)(pDest-4UL)) = PacketLength;
		}
	}
	/*
	 *	VLAN TAG Insertion Checking Ends
	 */
	else
	{
		/*
		 *	Copy the packet !
		 */ 

		SMSC_TRACE4(DBG_TX,"Copy pkt %p (size = %d) into TCB %p at offset %d\r\n", Packet, PacketLength, pTCB, pTCB->ulSize);

		// Set the Cmd A and B

		*((DWORD *) pDest) = TX_CMD_A_FIRST_SEG_ | TX_CMD_A_LAST_SEG_ | PacketLength;
		*((DWORD *)(pDest+4UL)) = PacketLength;
		pDest+= 8UL;
		pTCB->ulSize += 8UL;
		DestBufferSize -=8UL;
	}
	//
	// Copy the data from chained buffers to our TCB flat buffer.
	//
	while(CurrentBuffer)
	{
		NdisQueryBufferSafe(
			CurrentBuffer,
			&VirtualAddress,
			&CurrentLength,
			NormalPagePriority);
	
		if(!VirtualAddress) {
			SMSC_TRACE1(DBG_ERROR,"Could not get virtual address for Ndis buffer %p\r\n", CurrentBuffer);
			ASSERT(VirtualAddress != NULL);
			// This is an unrecoverable error for this packet, throw it away but signal 
			// caller to send out the previous ones we had already copied in this TCB.
			pTCB->ulSize = SaveInitialTcbDataSize;
			bResult = FALSE;
			goto Exit_NICCopyPacket;
		}
 	  
		if (CurrentLength > DestBufferSize)
		{
			SMSC_TRACE3(DBG_ERROR,"Not enogh space in destination. Currentbuffer=%p, CurrentbufferLength = %d, DestBufferLength = %d\r\n", CurrentBuffer, CurrentLength, DestBufferSize);
			// This is an unrecoverable error for this packet, throw it away but signal 
			// caller to send out the previous ones we had already copied in this TCB.
			pTCB->ulSize = SaveInitialTcbDataSize;
			bResult = FALSE;
			goto Exit_NICCopyPacket;
		}
		
		// Copy the data.
		NdisMoveMemory(pDest, VirtualAddress, CurrentLength);
		BytesCopied += CurrentLength;
		DestBufferSize -= CurrentLength;			
		pDest += CurrentLength;
		

		NdisGetNextBuffer(
			CurrentBuffer,
			&CurrentBuffer);
	}

	ASSERT (BytesCopied == BytesToCopy);

	Adapter->nSendBytesCopied += BytesCopied;
	Adapter->nSendPacketsCopied++;

	//
	// Adjust the buffer length to reflect the new length
	//
	pTCB->ulSize += BytesCopied;
	
Exit_NICCopyPacket:
	ASSERT (pTCB->ulSize <= Adapter->TxBurstMax);
	SMSC_TRACE0(DBG_TX,"<-- NICCopyPacket\r\n");
	return bResult;
}



/*++

Routine Description:

	This routine posts a TCB with data in it to the target device to send
	the network packet(s) out to the device.
		
Arguments:

	Adapter    - pointer to the MP_ADAPTER structure
	pTCB	-  Pointer to the TCB block that contains the IRP and data.


Return Value:

	NT status code

--*/
NTSTATUS
NICProcessTCB(
	PMP_ADAPTER Adapter,
	PTCB	pTCB
	)
{
	NTSTATUS ntStatus;

	SMSC_TRACE0(DBG_TX,"--> NICProcessTCB\r\n");

	NdisAcquireSpinLock(&Adapter->SendLock);

	if (Adapter->TxBatching)
	{
		if (pTCB == NULL)
		{
			pTCB = (PTCB) RemoveHeadList(&Adapter->SendTcbPendingList); 	   
			if(pTCB == (PTCB) &Adapter->SendTcbPendingList)
			{
				// No TCB here ... Must have sent it thru another thread
				// Anyway no worries, just release the lock and get out...
				NdisReleaseSpinLock(&Adapter->SendLock); 
	
				ntStatus = STATUS_SUCCESS;
				goto Exit_NICProcessTCB;	
			}

			// Got one TCB ! Update list counter and let it flow into 
			//	the send code below 		
			NdisInterlockedDecrement(&Adapter->nTcbPendingSend);
			ASSERT(Adapter->nTcbPendingSend >= 0);
		}
		
		else
		{

			if (pTCB->bFull != TRUE)
			{
				NdisInterlockedIncrement(&Adapter->nTcbPendingSend);
				InsertTailList(&Adapter->SendTcbPendingList, &pTCB->List);
				ASSERT(Adapter->nTcbPendingSend <= Adapter->MaxTCBs);
			
				if (Adapter->nTcbBusySend != 0)
				{
					// There's a TCB busy at the host controller, so 
					// leave it pending and exit
					NdisReleaseSpinLock(&Adapter->SendLock); 
	
					ntStatus = STATUS_PENDING;
					goto Exit_NICProcessTCB;	
				}

				pTCB = (PTCB) RemoveHeadList(&Adapter->SendTcbPendingList); 	   
			
				ASSERT(pTCB != (PTCB) &Adapter->SendTcbPendingList);
				if(pTCB == (PTCB) &Adapter->SendTcbPendingList)
				{
					// This could never happen, we just stuffed a tcb in the queue !!!
					SMSC_TRACE0(DBG_ERROR,"No TCB in the pending list when there should be!!!\r\n");
	
					NdisReleaseSpinLock(&Adapter->SendLock); 

					ntStatus = STATUS_UNSUCCESSFUL;
					goto Exit_NICProcessTCB;	
				}
			
				NdisInterlockedDecrement(&Adapter->nTcbPendingSend);
				ASSERT(Adapter->nTcbPendingSend >= 0);
			}
		}
		
	}
	
	if (!pTCB->ulSize)
	{

		NdisReleaseSpinLock(&Adapter->SendLock); 
		// Nothing to send, back to the free list
		// Do basically the same thing that NICFreeSendTCB() does
		// but without assuming the packet is in the busy queue
		// and do it while we hold the send lock 
		SMSC_TRACE1(DBG_TX,"--> NICFreeSendTCB inline subset %p\r\n", pTCB);
		pTCB->OrgSendPacket = NULL;
		pTCB->bFull=FALSE;
		pTCB->nPktCount = 0;
	

		ASSERT(!pTCB->Ref);

		NdisAcquireSpinLock(&Adapter->SendLock);
		InsertTailList(&Adapter->SendTcbFreeList, &pTCB->List); 	
		SMSC_TRACE1(DBG_TX,"<-- NICFreeSendTCB inline subset %p\r\n", pTCB);
		NdisReleaseSpinLock(&Adapter->SendLock); 

		ntStatus = STATUS_UNSUCCESSFUL;
		goto Exit_NICProcessTCB;	
	}

	NdisInterlockedIncrement(&Adapter->nTcbBusySend);
	ASSERT(Adapter->nTcbBusySend <= Adapter->MaxTCBs);

	InsertTailList(&Adapter->SendTcbBusyList, &pTCB->List);

	NdisReleaseSpinLock(&Adapter->SendLock); 

	Adapter->nBulkOutPosted++;

	//
	// Post a asynchronouse write request.
	//
	ntStatus = NICPostWriteRequest(Adapter, pTCB);
   
	if ( ntStatus ) {
		if(NdisInterlockedDecrement(&pTCB->Ref) == 0)
		{
			
			NICFreeSendTCB(Adapter, pTCB);
		}
		else 
		{
			// This should never happen
			SMSC_TRACE0(DBG_ERROR,"NICProcessTCB finds Pending TCB with invalid Ref count!\r\n");				 
		}
	} 

Exit_NICProcessTCB:
	SMSC_TRACE1(DBG_TX,"<-- NICProcessTCB (Status = 0x%08lx)\r\n", ntStatus);

	return ntStatus;
}


/*++

Routine Description:

	Completion routine for the write request. This routine
	put the TCBs into SendTCBCompletionList and shedule a workItem to proccess
	
Arguments:

   
	pTCB -	pointer to the context structure (TCB)
	
Return Value:

	STATUS_SUCCESS
--*/

NTSTATUS
NICWriteRequestCompletion (
	PTCB			pTCB
	)
{

	PMP_ADAPTER 	Adapter = pTCB->Adapter;
	
	SMSC_TRACE0(DBG_TX,"--> NICWriteRequestCompletion\r\n");

	NdisAcquireSpinLock(&Adapter->SendLock);
			
	// Remove it from the busy list 
	RemoveEntryList(&pTCB->List);	
	NdisInterlockedDecrement(&Adapter->nTcbBusySend);
	ASSERT(Adapter->nTcbBusySend >= 0);
			
	// Put it in the SendTcbcompletionList.
	InsertTailList(&Adapter->SendTcbcompletionList, &pTCB->List);
	NdisInterlockedIncrement(&Adapter->nTcbCompletionSend);
	ASSERT(Adapter->nTcbCompletionSend <= Adapter->MaxTCBs);
			
	SMSC_TRACE1(DBG_TX,"TCB %p removed from it's current list (busy if in any) and inserted into BulkInCancelledList\r\n", pTCB);
			
	NdisReleaseSpinLock(&Adapter->SendLock); 

	#ifdef USE_THREAD
		SetEvent(Adapter->SendThreadEvent);
	#else
		#ifdef USE_WORK_ITEM_NO_OUT
		 NdisSetEvent(&Adapter->SendWorkItemReady);
		#else
	
		if (InterlockedExchange(&Adapter->IsSendsCompletionWorkItemQueued, TRUE) == FALSE)
		{
			Adapter->nSendsCompletionWorkItemScheduled++;
			MP_INC_REF(Adapter);
			NdisScheduleWorkItem(&Adapter->SendsCompletionWorkItem);
		}
		#endif
	#endif
		
	SMSC_TRACE0(DBG_TX,"<-- NICWriteRequestCompletion\r\n");
	
	 return STATUS_SUCCESS;
}


/*++

Routine Description:

	This WorkItem is used to process SendTcbcompletionList
		
Arguments:

 
	Context -	pointer to the context structure (TCB) 

Return Value:

	STATUS_SUCCESS
	
--*/

NTSTATUS
SendsCompletionWorkItemCallBack(
	PNDIS_WORK_ITEM WorkItem, 
	PVOID Context
	)
{
	PMP_ADAPTER 	Adapter = Context;
	PTCB				pTCB;
	ULONG			bytesWritten;
	PNDIS_PACKET	Packet = NULL;
	USB_TRANSFER   hTransfer;
	USB_ERROR UsbErr;
	BOOL dwErr=FALSE;

	UNREFERENCED_PARAMETER(WorkItem);
	
	SMSC_TRACE0(DBG_TX,"--> SendsCompletionWorkItemCallBack\r\n");
		
	NdisWaitEvent(&Adapter->IsSendsCompletionWorkItemReady, 0);
	NdisResetEvent(&Adapter->IsSendsCompletionWorkItemReady);

	InterlockedExchange(&Adapter->IsSendsCompletionWorkItemQueued, FALSE);
	
	 NdisAcquireSpinLock(&Adapter->SendLock);
	while (!IsListEmpty(&Adapter->SendTcbcompletionList) ) {	
			
	  
		// Remove it from the SendTcbcompletionList
		pTCB = (PTCB) RemoveHeadList(&Adapter->SendTcbcompletionList);
		ASSERT(pTCB && (pTCB != (PTCB) &Adapter->SendTcbcompletionList));
		
		NdisInterlockedDecrement(&Adapter->nTcbCompletionSend);
		ASSERT(Adapter->nTcbCompletionSend >= 0);
			
		NdisReleaseSpinLock(&Adapter->SendLock); 

		//Sometimes the completion routine is called before the transfer complete. We need to sync. it here.
		//Wait for the Bulk out transfer to complete, so we can process a valid transfer handler
		WaitForSingleObject(pTCB->SyncEvent, 100);
		ResetEvent(pTCB->SyncEvent);
	 
		if (pTCB->dwUsbErr == 0) {
			SMSC_TRACE1(DBG_ERROR,"WriteRequest has NULL Transfer,	hTransfer=0x%x\r\n", pTCB->dwUsbErr);   
		}
	
		hTransfer=(USB_TRANSFER)pTCB->dwUsbErr;
		dwErr=GetTransferStatus(Adapter->UsbFuncs,hTransfer, &pTCB->dwTransferLength,&UsbErr);

		if (hTransfer){
			CloseTransferHandle(Adapter->UsbFuncs, hTransfer);
		}
		

		if ( dwErr	&&	(USB_NO_ERROR == UsbErr)  &&  (!Adapter->BulkOut.PipeTransferAborted) && hTransfer) 
		{  
			Adapter->nBulkOutCompletedSuccessfully++;
			Adapter->nPacketsTransmittedSuccessfully += (ULONG64)pTCB->nPktCount; 
			bytesWritten = pTCB->dwTransferLength;
			SMSC_TRACE1(DBG_TX,"Wrote %d bytes\r\n", bytesWritten); 
			Adapter->nBulkOutBytesWritten += bytesWritten;
		 } 
		else 
		{
			Adapter->nBulkOutCompletedWithError += pTCB->nPktCount;   
			SMSC_TRACE4(DBG_ERROR,"WriteRequest failed dwErr = 0x%x, UsbErr = 0x%x, Aborted= %d, hTransfer=WF0x%x\r\n", 
					dwErr,
					UsbErr,
					Adapter->BulkOut.PipeTransferAborted,
					hTransfer); 	   

		
			MP_CLEAR_FLAG(Adapter, fMP_POST_WRITES);

		
			// Handle USB Error recovery
			if ((UsbErr != USB_CANCELED_ERROR) &&
				(!Adapter->BulkOut.PipeTransferAborted) &&\
				(!MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS)) && \
				(!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS)) && \
				(!MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED)) && \
				(!MP_TEST_FLAG(Adapter, fMP_DEVICE_NOT_ATTACHED)) )
			 {
					SMSC_TRACE2(DBG_ERROR,"NICWriteRequestCompletion. USB Error !!! dwErr= %d, UsbErr = 0x%08lx\r\n", dwErr,UsbErr); 

					if (InterlockedExchange(&Adapter->IsUsbErrorRecoveryWorkItemQueued, TRUE) == FALSE) 
					{
							Adapter->nUsbErrorRecoveryWorkItemScheduled++;
							MP_INC_REF(Adapter);				   
							NdisScheduleWorkItem(&Adapter->UsbErrorRecoveryWorkItem);
					}
			}

		}

			
		if(NdisInterlockedDecrement(&pTCB->Ref) == 0)
		{
			 //
			 // For performance, let us free the TCB block before completing send.
			 //
			Packet = pTCB->OrgSendPacket;	// Only makes sense for single packet mode
			
			NICFreeCompletionTCB(Adapter, pTCB);			

			if(Adapter->DelayTxComplete)
			{
				SMSC_TRACE0(DBG_TX,"Calling NdisMSendComplete \r\n");
				Adapter->nSendPacketsCompleted++;			
				NdisMSendComplete(
							Adapter->AdapterHandle,
							Packet,
							NDIS_STATUS_SUCCESS);
			}

		}
		else 
		{
			// This should never happen
			SMSC_TRACE0(DBG_ERROR,"NICWriteRequestCompletion finds TCB with invalid Ref count!\r\n");			 
		}


		
		if (!IsListEmpty(&Adapter->SendTcbPendingList) && (Adapter->nTcbBusySend == 0))
			{
				// If so, schedule a workitem (do not call NICProcessTCB from here to avoid 
				// recursion if this completion was synchronous.
		
				SMSC_TRACE0(DBG_TX,"call NICProcessTCB\r\n");

				NICPostSendsWaitingCallBack(Adapter);
			}
			else
			{
				//
				// Before we exit, since we have the control, let us see if there any 
				// more packets waiting in the queues to be sent.
				//
				if (!IsListEmpty(&Adapter->SendPktWaitList))
				{
					// If so, schedule a workitem (do not call MPSend from here to avoid 
					// recursion if this completion was synchronous.
		
					NICPostSendsWaitingCallBack(Adapter);
				}
			}
		 NdisAcquireSpinLock(&Adapter->SendLock);

	}
	NdisReleaseSpinLock(&Adapter->SendLock); 

	NdisSetEvent(&Adapter->IsSendsCompletionWorkItemReady);
	MP_DEC_REF(Adapter);
	SMSC_TRACE0(DBG_TX,"<-- SendsCompletionWorkItemCallBack\r\n");
	return STATUS_SUCCESS;
}



/*++

Routine Description:

	This WorkItem is used to process SendTcbcompletionList
		
Arguments:

 
	Context -	pointer to the context structure (TCB) 

Return Value:

	STATUS_SUCCESS
	
--*/

NTSTATUS
SendsCompletionCallBack(
	PMP_ADAPTER 	Adapter
	)
{
	PTCB				pTCB;
	ULONG			bytesWritten;
	PNDIS_PACKET	Packet = NULL;
	USB_TRANSFER   hTransfer;
	USB_ERROR UsbErr;
	BOOL dwErr=FALSE;
	
	SMSC_TRACE0(DBG_TX,"--> SendsCompletionWorkItemCallBack\r\n");

	if (Adapter->MPInitializeFail ||(!MP_TEST_FLAG(Adapter, fMP_SEND_SIDE_RESOURCE_ALLOCATED)))
			return STATUS_SUCCESS;
		
	
	 NdisAcquireSpinLock(&Adapter->SendLock);
	while (!IsListEmpty(&Adapter->SendTcbcompletionList) ) {	
			
		
		// Remove it from the SendTcbcompletionList
		pTCB = (PTCB) RemoveHeadList(&Adapter->SendTcbcompletionList);
		ASSERT(pTCB && (pTCB != (PTCB) &Adapter->SendTcbcompletionList));
		
		NdisInterlockedDecrement(&Adapter->nTcbCompletionSend);
		ASSERT(Adapter->nTcbCompletionSend >= 0);
			
		NdisReleaseSpinLock(&Adapter->SendLock); 

		//Sometimes the completion routine is called before the transfer complete. We need to sync. it here.
		//Wait for the Bulk out transfer to complete, so we can process a valid transfer handler
		WaitForSingleObject(pTCB->SyncEvent, 100);
		ResetEvent(pTCB->SyncEvent);
	 
		if (pTCB->dwUsbErr == 0) {
			SMSC_TRACE1(DBG_ERROR,"WriteRequest has NULL Transfer,	hTransfer=0x%x\r\n", pTCB->dwUsbErr);   
		}
	
		hTransfer=(USB_TRANSFER)pTCB->dwUsbErr;
		dwErr=GetTransferStatus(Adapter->UsbFuncs,hTransfer, &pTCB->dwTransferLength,&UsbErr);

		if (hTransfer){
			CloseTransferHandle(Adapter->UsbFuncs, hTransfer);
		}
		

		if ( dwErr	&&	(USB_NO_ERROR == UsbErr)  &&  (!Adapter->BulkOut.PipeTransferAborted) && hTransfer) 
		{  
			Adapter->nBulkOutCompletedSuccessfully++;
			Adapter->nPacketsTransmittedSuccessfully += pTCB->nPktCount; 
			bytesWritten = pTCB->dwTransferLength;
			SMSC_TRACE1(DBG_TX,"Wrote %d bytes\r\n", bytesWritten); 
			Adapter->nBulkOutBytesWritten += bytesWritten;
		 } 
		else 
		{
			Adapter->nBulkOutCompletedWithError += pTCB->nPktCount;   
			SMSC_TRACE4(DBG_ERROR,"WriteRequest failed dwErr = 0x%x, UsbErr = 0x%x, Aborted= %d, hTransfer=WF0x%x\r\n", 
					dwErr,
					UsbErr,
					Adapter->BulkOut.PipeTransferAborted,
					hTransfer); 	   

		
			MP_CLEAR_FLAG(Adapter, fMP_POST_WRITES);

		
			// Handle USB Error recovery
			if ((UsbErr != USB_CANCELED_ERROR) &&
				(!Adapter->BulkOut.PipeTransferAborted) &&\
				(!MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS)) && \
				(!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS)) && \
				(!MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED)) && \
				(!MP_TEST_FLAG(Adapter, fMP_DEVICE_NOT_ATTACHED)) )
			 {
					SMSC_TRACE2(DBG_ERROR,"NICWriteRequestCompletion. USB Error !!! dwErr= %d, UsbErr = 0x%08lx\r\n", dwErr,UsbErr); 

					if (InterlockedExchange(&Adapter->IsUsbErrorRecoveryWorkItemQueued, TRUE) == FALSE) 
					{
							Adapter->nUsbErrorRecoveryWorkItemScheduled++;
							MP_INC_REF(Adapter);				   
							NdisScheduleWorkItem(&Adapter->UsbErrorRecoveryWorkItem);
					}
			}

		}

			
		if(NdisInterlockedDecrement(&pTCB->Ref) == 0)
		{
			 //
			 // For performance, let us free the TCB block before completing send.
			 //
			Packet = pTCB->OrgSendPacket;	// Only makes sense for single packet mode
			
			NICFreeCompletionTCB(Adapter, pTCB);			

			if(Adapter->DelayTxComplete)
			{
				SMSC_TRACE0(DBG_TX,"Calling NdisMSendComplete \r\n");
				Adapter->nSendPacketsCompleted++;			
				NdisMSendComplete(
							Adapter->AdapterHandle,
							Packet,
							NDIS_STATUS_SUCCESS);
			}

		}
		else 
		{
			// This should never happen
			SMSC_TRACE0(DBG_ERROR,"NICWriteRequestCompletion finds TCB with invalid Ref count!\r\n");			 
		}


		
		if (!IsListEmpty(&Adapter->SendTcbPendingList) && (Adapter->nTcbBusySend == 0))
			{
				// If so, schedule a workitem (do not call NICProcessTCB from here to avoid 
				// recursion if this completion was synchronous.
		
				SMSC_TRACE0(DBG_TX,"call NICProcessTCB\r\n");

				NICPostSendsWaitingCallBack(Adapter);
			}
			else
			{
				//
				// Before we exit, since we have the control, let us see if there any 
				// more packets waiting in the queues to be sent.
				//
				if (!IsListEmpty(&Adapter->SendPktWaitList))
				{
					// If so, schedule a workitem (do not call MPSend from here to avoid 
					// recursion if this completion was synchronous.
		
					NICPostSendsWaitingCallBack(Adapter);
				}
			}
		 NdisAcquireSpinLock(&Adapter->SendLock);

	}
	NdisReleaseSpinLock(&Adapter->SendLock); 

	SMSC_TRACE0(DBG_TX,"<-- SendsCompletionWorkItemCallBack\r\n");
	return STATUS_SUCCESS;
}


/*++

Routine Description:

		This routine cancels any pending sends not submitted to USB.
Arguments:

		Adapter - pointer to our adapter context structure

Return Value:
	None
	
--*/
VOID
NICFreePendingSendPackets(
	PMP_ADAPTER Adapter
	)
{ 
	PTCB				pTCB = NULL;
	
	SMSC_TRACE0(DBG_TX,"--> NICFreeBusySendPackets\r\n");

	if(!MP_TEST_FLAG(Adapter, fMP_SEND_SIDE_RESOURCE_ALLOCATED)){
		return;
	}		 

	NdisAcquireSpinLock(&Adapter->SendLock);
	while(!IsListEmpty(&Adapter->SendTcbPendingList))
	{
		pTCB = (PTCB) RemoveHeadList(&Adapter->SendTcbPendingList);
		NdisInterlockedDecrement(&Adapter->nTcbPendingSend);
		//
		// Initialize the head so that we don't barf in NICFreeSendTCB
		// while doing RemoveEntryList.
		//
		NdisInitializeListHead(&pTCB->List);					

		// "Virtually" put it in the Tcb Busy List (not really there
		// but the NICFreeSendTCB() call below will decrement this count
		NdisInterlockedIncrement(&Adapter->nTcbBusySend);

		NdisReleaseSpinLock(&Adapter->SendLock); 
		
		if(NdisInterlockedDecrement(&pTCB->Ref) == 0)
		{
			if (Adapter->DelayTxComplete)
			{
				// Only makes sense for single packet mode
				SMSC_TRACE0(DBG_TX,"Calling NdisMSendComplete \r\n");

				Adapter->nSendPacketsCompleted++;
			
				NdisMSendComplete(Adapter->AdapterHandle,
									pTCB->OrgSendPacket,
									NDIS_STATUS_SUCCESS); 
			}
			NICFreeSendTCB(Adapter, pTCB);
		}
		else 
		{
			// This should never happen
			SMSC_TRACE0(DBG_ERROR,"NICFreeBusySendPackets finds Pending TCB with invalid Ref count!\r\n");				 
		}

		NdisAcquireSpinLock(&Adapter->SendLock);
	}
	
	NdisReleaseSpinLock(&Adapter->SendLock); 

	SMSC_TRACE0(DBG_TX,"<-- NICFreeBusySendPackets\r\n");
	 
	return ;
}



